#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..ecdsa.enums.Ecc import Ecc
from ..asn.enums.X9ObjectIdentifier import X9ObjectIdentifier
from ..asn.enums.CertificateVersion import CertificateVersion
from ..internal._GXCommon import _GXCommon
from ..asn.GXAsn1Converter import GXAsn1Converter
from ..asn.X9ObjectIdentifierConverter import X9ObjectIdentifierConverter
from .GXPublicKey import GXPublicKey
from ..GXBitString import GXBitString
from ..asn.GXAsn1Sequence import GXAsn1Sequence
from ..asn.GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier
from ..asn.GXAsn1Context import GXAsn1Context
from .GXCurve import GXCurve
from .GXEccPoint import GXEccPoint
from ._GXShamirs import _GXShamirs
from ..GXByteBuffer import GXByteBuffer
from ..GXInt8 import GXInt8

# from gurux_dlms.GXDLMSTranslator import GXDLMSTranslator


class GXPrivateKey:
    """
    Private key.
    """

    __scheme = Ecc.P256
    """
    Used scheme.
    """

    __rawValue = bytearray()
    """
    Private key raw value.
    """

    __systemTitle = bytearray()
    """
    SystemTitle is an extra information that can be used in debugging.
    """

    __publicKey = None

    @property
    def scheme(self):
        """
        Used scheme.
        """
        return self.__scheme

    @property
    def rawValue(self):
        """
        Private key raw value.
        """
        return self.__rawValue

    @property
    def systemTitle(self):
        """
        SystemTitle is an extra information that can be used in debugging.
        """
        return self.__systemTitle

    @systemTitle.setter
    def systemTitle(self, value):
        self.__systemTitle = value

    # pylint: disable=unused-private-member
    @classmethod
    def fromRawBytes(cls, key):
        """
        Create the private key from raw bytes.

            Parameters:
                key: Raw data

            Returns:
                Private key.
        """
        value = GXPrivateKey()
        # If private key is given
        if len(key) == 32:
            value.__scheme = Ecc.P256
            value.__rawValue = key
        elif len(key) == 48:
            value.__scheme = Ecc.P384
            value.__rawValue = key
        else:
            raise ValueError("Invalid private key.")
        return value

    # pylint: disable=unused-private-member
    @classmethod
    def fromDer(cls, der):
        """
        Create the private key from DER.

            Parameters:
                key: DER Base64 coded string.

            Returns:
                Private key.
        """
        der = der.replace("\r\n", "")
        der = der.replace("\n", "")
        key = _GXCommon.fromBase64(der)
        seq = GXAsn1Converter.fromByteArray(key)
        if seq[0] > 3:
            raise ValueError("Invalid private key version.")
        tmp = seq[2]
        value = GXPrivateKey()
        id_ = X9ObjectIdentifierConverter.fromString(str(tmp[0]))
        if id_ == X9ObjectIdentifier.PRIME256V1:
            value.__scheme = Ecc.P256
        elif id_ == X9ObjectIdentifier.SECP384R1:
            value.__scheme = Ecc.P384
        else:
            if id_ == X9ObjectIdentifier.NONE:
                raise ValueError("Invalid private key " + str(tmp[0]) + ".")
            raise ValueError("Invalid private key " + str(id_) + " " + str(tmp[0]) + ".")

        value.__rawValue = seq[1]
        if isinstance(seq[3], bytes):
            value.__publicKey = GXPublicKey.fromRawBytes(seq[3])
        elif isinstance(seq[3], GXBitString):
            value.__publicKey = GXPublicKey.fromRawBytes((seq[3]).value)
        else:
            # Open SSL PEM.
            value.__publicKey = GXPublicKey.fromRawBytes(seq[3][0].value)
        return value

    @classmethod
    def fromPem(cls, pem):
        """
        Create the private key from PEM.

            Parameters:
                pem: PEM in Base64 coded string.

        Returns
            Private key.
        """
        pem = pem.replace("\r\n", "\n")
        START = "PRIVATE KEY-----\n"
        END = "-----END"
        index = pem.find(START)
        if index == -1:
            raise ValueError("Invalid PEM file.")
        pem = pem[index + len(START) :]
        index = pem.find(END)
        if index == -1:
            raise ValueError("Invalid PEM file.")
        return cls.fromDer(pem[0:index])

    @classmethod
    def load(cls, path):
        """
        Create the private key from PEM file.

            Parameters:
                path: Path to the PEM file.

            Returns
                Private key.
        """
        with open(path, "r", encoding="utf-8") as file:
            return cls.fromPem(file.read())

    def save(self, path):
        """
        Save private key to PEM file.

            Parameters:
                path: File path.
        """
        with open(path, "w", encoding="utf-8") as file:
            file.write(self.toPem())

    def toDer(self):
        d = GXAsn1Sequence()
        d.append(GXInt8(CertificateVersion.VERSION2))
        d.append(self.__rawValue)
        d1 = GXAsn1Sequence()
        if self.scheme == Ecc.P256:
            d1.append(GXAsn1ObjectIdentifier("1.2.840.10045.3.1.7"))
        elif self.scheme == Ecc.P384:
            d1.append(GXAsn1ObjectIdentifier("1.3.132.0.34"))
        else:
            raise ValueError("Invalid ECC scheme.")
        d.append(d1)
        d2 = GXAsn1Context()
        d2.index = 1
        d2.append(GXBitString(self.getPublicKey().rawValue, 0))
        d.append(d2)
        return _GXCommon.toBase64(GXAsn1Converter.toByteArray(d))

    def toPem(self):
        """
        Get private key as PEM format.
        """
        return (
            "-----BEGIN EC PRIVATE KEY-----\n"
            + self.toDer()
            + "\n-----END EC PRIVATE KEY-----"
        )

    def getPublicKey(self):
        """
        Get public key from private key.

            @returns:
                Public key.
        """
        if not self.__publicKey:
            pk = int.from_bytes(self.__rawValue, byteorder="big")
            curve = GXCurve(self.scheme)
            ret = GXEccPoint()
            _GXShamirs.pointMulti(curve, ret, curve.g, pk)
            if self.scheme == Ecc.P256:
                size = 32
            else:
                size = 48
            # Public key is un-compressed format.
            key = bytearray(int(4).to_bytes(1))
            key.extend(ret.x.to_bytes(size, byteorder="big"))
            key.extend(ret.y.to_bytes(size, byteorder="big"))
            self.__publicKey = GXPublicKey.fromRawBytes(key)
        return self.__publicKey

    def toHex(self, addSpace=True):
        """
           Returns the private key as a hex string.

               Parameters:
                   addSpace: Is space added between the bytes.

        *
        * @return Private key as hex string.
        """
        return GXByteBuffer.hex(self.__rawValue, addSpace)

    def __eq__(self, other):
        return isinstance(other, GXPrivateKey) and self.__rawValue == other.rawValue

    def __str__(self):
        return self.toHex(True)



#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import os

from .enums.CertificateVersion import CertificateVersion
from .enums.ExtendedKeyUsage import ExtendedKeyUsage
from .enums.HashAlgorithm import HashAlgorithm
from .enums.KeyUsage import KeyUsage
from ..enums.BerType import BerType

from ..objects.enums.CertificateType import CertificateType
from ..internal._GXCommon import _GXCommon
from ..ecdsa.enums.Ecc import Ecc
from ..GXByteBuffer import GXByteBuffer
from .HashAlgorithmConverter import HashAlgorithmConverter
from .GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier
from ..GXBitString import GXBitString
from ..ecdsa.GXEcdsa import GXEcdsa
from ..ecdsa.GXDLMSCertificateException import GXDLMSCertificateException
from .GXAsn1Sequence import GXAsn1Sequence
from .GXAsn1Integer import GXAsn1Integer
from .X509CertificateTypeConverter import X509CertificateTypeConverter
from .enums.X509CertificateType import X509CertificateType
from .X509NameConverter import X509NameConverter
from .enums.X509Name import X509Name
from ..GXArray import GXArray
from .GXAsn1Context import GXAsn1Context
from ..GXInt8 import GXInt8
from .enums.PkcsType import PkcsType


# pylint: disable=too-many-public-methods, too-many-instance-attributes
class GXx509Certificate:
    """
    x509 Certificate.
    """

    __description = ""
    """
    Description is extra metadata that is saved to PEM file.
    """

    __rawData = bytearray()
    """
    Loaded x509Certificate as raw data.
    """

    __subjectKeyIdentifier = bytearray()
    """
    This extension identifies the public key being certified.
    """

    __authorityKeyIdentifier = bytearray()
    """
    May be used either as a certificate or CRL extension. It identifies the
        
 public key to be used to verify the signature on this certificate or CRL.
        
 It enables distinct keys used by the same CA to be distinguished.
    """

    __authorityCertificationSerialNumber = None
    """
    Authority certification serial number.
    """

    __basicConstraints = None
    """
    Indicates if the Subject may act as a CA.
    """

    __extendedKeyUsage = ExtendedKeyUsage.NONE
    """
    Indicates that a certificate can be used as an TLS server or client certificate.
    """

    __signatureAlgorithm = HashAlgorithm.SHA256WITH_ECDSA
    """
    Signature algorithm.
    """

    __signatureParameters = None
    """
    Signature Parameters.
    """

    __publicKey = None
    """
    Public key.
    """

    __publicKeyAlgorithm = HashAlgorithm.NONE
    """
    Public Key algorithm.
    """

    __publicKeyParameters = None
    """
    Parameters.
    """

    __signature = bytearray()
    """
    Signature.
    """

    __subject = ""
    """
    Subject. Example: "CN=Test, O=Gurux, L=Tampere, C=FI".
    """

    __subjectAlternativeName = ""
    """
    Subject Alternative Name.
    """

    __issuer = ""
    """
    Issuer. Example: "CN=Test O=Gurux, L=Tampere, C=FI".
    """

    __issuerRaw = bytearray()
    """
    Raw Issuer in ASN1 format.
    """

    __authorityCertIssuer = ""
    """
    Authority Cert Issuer. Example: "CN=Test O=Gurux, L=Tampere, C=FI".
    """

    __serialNumber = None
    """
    Serial number.
    """

    __version = CertificateVersion.VERSION3
    """
    Version.
    """

    __validFrom = None
    """
    Validity from.
    """

    __validTo = None
    """
    Validity to.
    """

    __keyUsage = KeyUsage.NONE
    """
    Indicates the purpose for which the certified public key is used.
    """

    # pylint: disable=import-outside-toplevel
    @classmethod
    def getFilePath(cls, cert):
        """Returns the default file path."""
        from .GXAsn1Converter import GXAsn1Converter

        if cert.keyUsage == KeyUsage.DIGITAL_SIGNATURE:
            path = "D"
        elif cert.keyUsage == KeyUsage.KEY_AGREEMENT:
            path = "A"
        elif cert.keyUsage == (KeyUsage.DIGITAL_SIGNATURE | KeyUsage.KEY_AGREEMENT):
            path = "T"
        else:
            raise ValueError("Unknown certificate type.")

        st = GXAsn1Converter.hexSystemTitleFromSubject(cert.subject).strip()
        path += st + ".pem"
        if cert.publicKey.scheme == Ecc.P256:
            path = os.path.join("Certificates", path)
        else:
            path = os.path.join("Certificates384", path)
        return path

    @property
    def description(self):
        """
        Description is extra metadata that is saved to PEM file.
        """
        return self.__description

    @description.setter
    def description(self, value):
        self.__description = value

    @property
    def rawData(self):
        """
        Loaded x509Certificate as raw data.
        """
        return self.__rawData

    @property
    def subjectKeyIdentifier(self):
        """
        This extension identifies the public key being certified.
        """
        return self.__subjectKeyIdentifier

    @subjectKeyIdentifier.setter
    def subjectKeyIdentifier(self, value):
        self.__subjectKeyIdentifier = value

    @property
    def authorityKeyIdentifier(self):
        """
        May be used either as a certificate or CRL extension. It identifies the
        public key to be used to verify the signature on this certificate or CRL.
        It enables distinct keys used by the same CA to be distinguished.
        """
        return self.__authorityKeyIdentifier

    @authorityKeyIdentifier.setter
    def authorityKeyIdentifier(self, value):
        self.__authorityKeyIdentifier = value

    @property
    def authorityCertificationSerialNumber(self):
        """
        Authority certification serial number.
        """
        return self.__authorityCertificationSerialNumber

    @authorityCertificationSerialNumber.setter
    def authorityCertificationSerialNumber(self, value):
        self.__authorityCertificationSerialNumber = value

    @property
    def basicConstraints(self):
        """
        Indicates if the Subject may act as a CA.
        """
        return self.__basicConstraints

    @basicConstraints.setter
    def basicConstraints(self, value):
        self.__basicConstraints = value

    @property
    def extendedKeyUsage(self):
        """
        Indicates that a certificate can be used as an TLS server or client certificate.
        """
        return self.__extendedKeyUsage

    @extendedKeyUsage.setter
    def extendedKeyUsage(self, value):
        self.__extendedKeyUsage = value

    @property
    def signatureAlgorithm(self):
        """
        Signature algorithm.
        """
        return self.__signatureAlgorithm

    @signatureAlgorithm.setter
    def signatureAlgorithm(self, value):
        self.__signatureAlgorithm = value

    @property
    def signatureParameters(self):
        """
        Signature Parameters.
        """
        return self.__signatureParameters

    @signatureParameters.setter
    def signatureParameters(self, value):
        self.__signatureParameters = value

    @property
    def publicKey(self):
        """
        Public key.
        """
        return self.__publicKey

    @publicKey.setter
    def publicKey(self, value):
        self.__publicKey = value

    @property
    def publicKeyAlgorithm(self):
        """
        Public Key algorithm.
        """
        return self.__publicKeyAlgorithm

    @publicKeyAlgorithm.setter
    def publicKeyAlgorithm(self, value):
        self.__publicKeyAlgorithm = value

    @property
    def publicKeyParameters(self):
        """
        Parameters.
        """
        return self.__publicKeyParameters

    @publicKeyParameters.setter
    def publicKeyParameters(self, value):
        self.__publicKeyParameters = value

    @property
    def signature(self):
        """
        Signature.
        """
        return self.__signature

    @signature.setter
    def signature(self, value):
        self.__signature = value

    @property
    def subject(self):
        """
        Subject. Example: "CN=Test, O=Gurux, L=Tampere, C=FI".
        """
        return self.__subject

    @subject.setter
    def subject(self, value):
        self.__subject = value

    @property
    def subjectAlternativeName(self):
        """
        Subject Alternative Name.
        """
        return self.__subjectAlternativeName

    @subjectAlternativeName.setter
    def subjectAlternativeName(self, value):
        self.__subjectAlternativeName = value

    @property
    def issuer(self):
        """
        Issuer. Example: "CN=Test O=Gurux, L=Tampere, C=FI".
        """
        return self.__issuer

    @issuer.setter
    def issuer(self, value):
        self.__issuer = value

    @property
    def issuerRaw(self):
        """
        Raw Issuer in ASN1 format.
        """
        return self.__issuerRaw

    @property
    def authorityCertIssuer(self):
        """
        Authority Cert Issuer. Example: "CN=Test O=Gurux, L=Tampere, C=FI".
        """
        return self.__authorityCertIssuer

    @authorityCertIssuer.setter
    def authorityCertIssuer(self, value):
        self.__authorityCertIssuer = value

    @property
    def serialNumber(self):
        """
        Serial number.
        """
        return self.__serialNumber

    @serialNumber.setter
    def serialNumber(self, value):
        self.__serialNumber = value

    @property
    def version(self):
        """
        Version.
        """
        return self.__version

    @property
    def validFrom(self):
        """
        Validity from.
        """
        return self.__validFrom

    @validFrom.setter
    def validFrom(self, value):
        self.__validFrom = value

    @property
    def validTo(self):
        """
        Validity to.
        """
        return self.__validTo

    @validTo.setter
    def validTo(self, value):
        self.__validTo = value

    @property
    def keyUsage(self):
        """
        Indicates the purpose for which the certified public key is used.
        """
        return self.__keyUsage

    @keyUsage.setter
    def keyUsage(self, value):
        self.__keyUsage = value

    def __init__(self, data=None):
        """
        Constructor.
        """
        if data:
            self.__init(data)

    @classmethod
    def fromHexString(cls, data):
        """
        Create x509Certificate from hex string.

            Parameters:
                data: Hex string.
            Returns:
                x509 certificate
        """
        cert = GXx509Certificate()
        cert.__init(GXByteBuffer.hexToBytes(data))
        return cert

    @classmethod
    def fromPem(cls, data):
        """
        Create x509Certificate from PEM string.

            Parameters:
                data: PEM string.

            Returns:
                x509 certificate
        """
        data = data.replace("\r\n", "\n")
        START = "CERTIFICATE-----\n"
        END = "-----END"
        data = data.replace("\r\n", "\n")
        start = data.find(START)
        if start == -1:
            raise ValueError("Invalid PEM file.")
        desc = None
        if start != 11:
            desc = data[:start]
            DESCRIPTION = "#Description"
            # Check if there is a description metadata.
            descStart = desc.rfind(DESCRIPTION)
            if descStart != -1:
                descEnd = desc.find("\n", descStart, start)
                desc = desc[descStart + len(DESCRIPTION), descEnd - len(DESCRIPTION)]
                desc = desc.strip()
            else:
                desc = None

        data = data[start + len(START) :]
        end = data.find(END)
        if end == -1:
            raise ValueError("Invalid PEM file.")
        cert = cls.fromDer(data[0:end])
        cert.description = desc
        return cert

    @classmethod
    def fromDer(cls, der):
        """
        Create x509Certificate from DER Base64 encoded string.

            Parameters:
                der: Base64 DER string.

            Returns:
                x509 certificate
        """
        der = der.replace("\r\n", "")
        der = der.replace("\n", "")
        cert = GXx509Certificate()
        cert.__init(_GXCommon.fromBase64(der))
        return cert

    # pylint: disable=import-outside-toplevel, too-many-locals, protected-access
    def __init(self, data):
        from .GXAsn1Converter import GXAsn1Converter
        from ..ecdsa.GXPublicKey import GXPublicKey

        self.__rawData = data
        seq = GXAsn1Converter.fromByteArray(data)
        if len(seq) != 3:
            raise GXDLMSCertificateException(
                "Invalid Certificate Version. Wrong number of elements in sequence."
            )

        if not isinstance(seq[0], GXAsn1Sequence):
            certType = GXAsn1Converter.__getCertificateType(data, seq)
            if certType == PkcsType.PKCS8:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is PKCS 8 private key, not x509 certificate."
                )
            if certType == PkcsType.PKCS10:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is PKCS 10 certification requests, not x509 certificate."
                )
            raise GXDLMSCertificateException("Invalid Certificate Version.")

        req_info = seq[0]

        if isinstance(req_info[0], GXAsn1Integer):
            raise GXDLMSCertificateException(
                "Invalid Certificate. DLMS certificate version number must be integer."
            )

        self.__version = CertificateVersion(req_info[0][0])
        self.__serialNumber = req_info[1]
        self.__signatureAlgorithm = HashAlgorithmConverter.fromString(
            str(req_info[2][0])
        )
        if self.__signatureAlgorithm not in [
            HashAlgorithm.SHA256WITH_ECDSA,
            HashAlgorithm.SHA384WITH_ECDSA,
        ]:
            raise Exception(
                "DLMS certificate must be signed with ECDSA with SHA256 or SHA384."
            )

        if len(req_info[2]) > 1:
            self.__signatureParameters = req_info[2][1]

        self.__issuerRaw = GXAsn1Converter.toByteArray(req_info[3])
        self.__issuer = GXAsn1Converter.getSubject(req_info[3])

        validity = req_info[4]

        self.__validFrom = validity[0]
        self.__validTo = validity[1]
        self.__subject = GXAsn1Converter.getSubject(req_info[5])

        subject_pk_info = req_info[6]
        self.__publicKey = GXPublicKey.fromRawBytes(subject_pk_info[1].value)
        GXEcdsa.validate(self.__publicKey)
        basicConstraintsExists = False
        if len(req_info) > 7:
            for s in req_info[7][0]:
                oid = s[0]
                value = s[1]
                cert_type = X509CertificateTypeConverter.fromString(str(oid))

                if cert_type == X509CertificateType.SUBJECT_KEY_IDENTIFIER:
                    self.subject_key_identifier = value

                elif cert_type == X509CertificateType.SUBJECT_ALTERNATIVE_NAME:
                    alt_names = []
                    for it in value:
                        alt_names.append("DNS:" + it[0].decode("ascii"))
                    self.subject_alternative_name = ", ".join(alt_names)

                elif cert_type == X509CertificateType.AUTHORITY_KEY_IDENTIFIER:
                    for it in value:
                        if it.index == 0:
                            self.authority_key_identifier = it[0]
                        elif it.index == 1:
                            pairs = [
                                f"{X509NameConverter.fromString(str(k))}={v}"
                                for k, v in it[0][0]
                            ]
                            self.authority_cert_issuer = ", ".join(pairs)
                        elif it.index == 2:
                            self.__authorityCertificationSerialNumber = int.from_bytes(
                                it[0], "big"
                            )
                        else:
                            raise Exception(f"Invalid context. {it.index}")

                elif cert_type == X509CertificateType.KEY_USAGE:
                    if isinstance(value, GXBitString):
                        self.__keyUsage = KeyUsage(value.toInteger())
                    else:
                        self.__keyUsage = KeyUsage(value[2].toInteger())

                elif cert_type == X509CertificateType.EXTENDED_KEY_USAGE:
                    self.__extendedKeyUsage = ExtendedKeyUsage.NONE
                    for it in value:
                        if it.object_identifier == "1.3.6.1.5.5.7.3.1":
                            self.__extendedKeyUsage |= ExtendedKeyUsage.SERVER_AUTH
                        elif it.object_identifier == "1.3.6.1.5.5.7.3.2":
                            self.__extendedKeyUsage |= ExtendedKeyUsage.CLIENT_AUTH
                        else:
                            raise Exception("Invalid extended key usage.")

                elif cert_type == X509CertificateType.BASIC_CONSTRAINTS:
                    basicConstraintsExists = True
                    if isinstance(value, GXAsn1Sequence) and len(value) > 0:
                        self.basic_constraints = bool(value[0])
                    else:
                        self.basic_constraints = bool(value)

        if not basicConstraintsExists:
            cnFound = False
            for k, v in req_info[5]:
                if (
                    X509NameConverter.getString(X509Name.CN) == str(k)
                    and len(str(v)) == 16
                ):
                    cnFound = True
                    break
            if not cnFound:
                raise GXDLMSCertificateException(
                    "Common name doesn't exist or system title is invalid."
                )

        if self.__keyUsage == KeyUsage.NONE:
            raise Exception("Key usage not present. It's mandatory.")

        if (
            self.__keyUsage & (KeyUsage.KEY_CERT_SIGN | KeyUsage.CRL_SIGN)
        ) and not basicConstraintsExists:
            raise Exception("Basic Constraints value not present. It's mandatory.")

        if (
            self.__keyUsage == (KeyUsage.DIGITAL_SIGNATURE | KeyUsage.KEY_AGREEMENT)
            and self.__extendedKeyUsage == ExtendedKeyUsage.NONE
        ):
            raise Exception("Extended key usage not present. It's mandatory for TLS.")

        if self.__extendedKeyUsage != ExtendedKeyUsage.NONE and self.__keyUsage != (
            KeyUsage.DIGITAL_SIGNATURE | KeyUsage.KEY_AGREEMENT
        ):
            raise Exception("Extended key usage present. It's used only for TLS.")

        self.__publicKeyAlgorithm = HashAlgorithmConverter.fromString(str(seq[1][0]))

        if self.__publicKeyAlgorithm not in [
            HashAlgorithm.SHA256WITH_ECDSA,
            HashAlgorithm.SHA384WITH_ECDSA,
        ]:
            raise Exception(
                "DLMS certificate must be signed with ECDSA with SHA256 or SHA384."
            )

        if len(seq[1]) > 1:
            self.__publicKeyParameters = seq[1][1]
        self.__signature = seq[2].value

    # pylint: disable=import-outside-toplevel,too-many-locals
    def getDataList(self):
        from .GXAsn1Converter import GXAsn1Converter

        if not self.issuer:
            raise ValueError("Issuer is empty.")
        if not self.subject:
            raise ValueError("Subject is empty.")

        a = GXAsn1ObjectIdentifier(
            HashAlgorithmConverter.getString(self.__signatureAlgorithm)
        )

        p = GXAsn1Context()
        p.append(GXInt8(self.version))

        s = GXAsn1Sequence()

        if self.__subjectKeyIdentifier:
            s1 = GXAsn1Sequence()
            s1.append(
                GXAsn1ObjectIdentifier(
                    X509CertificateTypeConverter.getString(
                        X509CertificateType.SUBJECT_KEY_IDENTIFIER
                    )
                )
            )
            bb = GXByteBuffer()
            bb.setUInt8(BerType.OCTET_STRING)
            _GXCommon.setObjectCount(len(self.__subjectKeyIdentifier), bb)
            bb.set(self.__subjectKeyIdentifier)
            s1.append(bb.array())
            s.append(s1)

            if (
                self.authorityKeyIdentifier
                or self.authorityCertIssuer
                or self.authorityCertificationSerialNumber
            ):
                s1 = GXAsn1Sequence()
                s1.append(
                    GXAsn1ObjectIdentifier(
                        X509CertificateTypeConverter.getString(
                            X509CertificateType.AUTHORITY_KEY_IDENTIFIER
                        )
                    )
                )
                s.append(s1)
                s2 = GXAsn1Context()
                s2.index = 3
                c1 = GXAsn1Sequence()
                if self.authorityKeyIdentifier:
                    c4 = GXAsn1Context()
                    c4.constructed = False
                    c4.index = 0
                    c4.append(self.authorityKeyIdentifier)
                    c1.append(c4)
                    s1.append(GXAsn1Converter.toByteArray(c1))

                if self.authorityCertIssuer:
                    c2 = GXAsn1Context()
                    c2.index = 1
                    c1.append(c2)
                    c3 = GXAsn1Context()
                    c3.index = 4
                    c2.append(c3)
                    c3.append(GXAsn1Converter.encodeSubject(self.authorityCertIssuer))
                    s2.append(c1)
                if self.authorityCertificationSerialNumber:
                    c4 = GXAsn1Context()
                    c4.constructed = False
                    c4.index = 2
                    tmp5 = self.authorityCertificationSerialNumber.toByteArray()
                    tmp5.reverse()
                    c4.append(tmp5)
                    c1.append(c4)
                    s1.append(GXAsn1Converter.toByteArray(c1))
        # BasicConstraints
        s1 = GXAsn1Sequence()
        s1.append(
            GXAsn1ObjectIdentifier(
                X509CertificateTypeConverter.getString(
                    X509CertificateType.BASIC_CONSTRAINTS
                )
            )
        )
        seq = GXAsn1Sequence()

        if self.__basicConstraints:
            s1.append(self.__basicConstraints)
        elif self.__keyUsage == KeyUsage.NONE:
            raise Exception("Key usage not present.")

        s1.append(GXAsn1Converter.toByteArray(seq))
        s.append(s1)

        # KeyUsage
        s1 = GXAsn1Sequence()
        s1.append(
            GXAsn1ObjectIdentifier(
                X509CertificateTypeConverter.getString(X509CertificateType.KEY_USAGE)
            )
        )

        keyUsage = _GXCommon.swapBits(self.__keyUsage)
        min_bit = min([bit for bit in KeyUsage if (bit & keyUsage)])
        ignore = min_bit.bit_length() - 1 if min_bit else 0

        bs = GXBitString([ignore % 8, keyUsage])
        s1.append(GXAsn1Converter.toByteArray(bs))
        s.append(s1)

        # ExtendedKeyUsage
        if self.__extendedKeyUsage != ExtendedKeyUsage.NONE:
            s1 = GXAsn1Sequence()
            s1.append(
                GXAsn1ObjectIdentifier(
                    X509CertificateTypeConverter.getString(
                        X509CertificateType.EXTENDED_KEY_USAGE
                    )
                )
            )
            s2 = GXAsn1Sequence()
            if self.__extendedKeyUsage & ExtendedKeyUsage.SERVER_AUTH:
                s2.append(GXAsn1ObjectIdentifier("1.3.6.1.5.5.7.3.1"))
            if self.__extendedKeyUsage & ExtendedKeyUsage.CLIENT_AUTH:
                s2.append(GXAsn1ObjectIdentifier("1.3.6.1.5.5.7.3.2"))

            s1.append(GXAsn1Converter.toByteArray(s2))
            s.append(s1)

        valid = GXAsn1Sequence()
        valid.append(self.__validFrom)
        valid.append(self.__validTo)
        if self.__publicKey.scheme == Ecc.P256:
            alg = GXAsn1ObjectIdentifier("1.2.840.10045.3.1.7")
        else:
            alg = GXAsn1ObjectIdentifier("1.3.132.0.34")
        tmp3 = GXArray((GXAsn1ObjectIdentifier("1.2.840.10045.2.1"), alg))
        tmp4 = GXAsn1Context()
        tmp4.index = 3
        tmp4.append(s)

        tmp2 = GXArray((tmp3, GXBitString(self.__publicKey.rawValue, 0)))
        if self.signatureParameters:
            p2 = GXArray((a, self.signatureParameters))
        else:
            p2 = GXArray((a,))

        return GXArray(
            (
                p,
                GXAsn1Integer(self.__serialNumber),
                p2,
                GXAsn1Converter.encodeSubject(self.__issuer),
                valid,
                GXAsn1Converter.encodeSubject(self.__subject),
                tmp2,
                tmp4,
            )
        )

    # pylint: disable=import-outside-toplevel
    @property
    def encoded(self):
        from .GXAsn1Converter import GXAsn1Converter

        if self.__rawData:
            return self.__rawData
        tmp = GXArray(
            (
                GXAsn1ObjectIdentifier(
                    HashAlgorithmConverter.getString(self.signatureAlgorithm)
                ),  # Don't remove the comma.
            )
        )
        list_ = GXArray((self.getDataList(), tmp, GXBitString(self.signature, 0)))
        self.__rawData = GXAsn1Converter.toByteArray(list_)
        return self.__rawData

    # pylint: disable=import-outside-toplevel
    def getData(self):
        """
        Get data as byte array.
        """
        from .GXAsn1Converter import GXAsn1Converter

        return GXAsn1Converter.toByteArray(self.getDataList())

    def __str__(self):
        if self.extendedKeyUsage == ExtendedKeyUsage.SERVER_AUTH:
            bb = "Server certificate" + os.linesep
        elif ExtendedKeyUsage == ExtendedKeyUsage.CLIENT_AUTH:
            bb = "Client certificate" + os.linesep
        elif ExtendedKeyUsage == (
            ExtendedKeyUsage.SERVER_AUTH | ExtendedKeyUsage.CLIENT_AUTH
        ):
            bb = "TLS certificate" + os.linesep
        bb += os.linesep
        bb += "Version: " + str(self.version)
        bb += os.linesep + "SerialNumber: " + str(self.serialNumber)
        bb += os.linesep + "Signature: " + str(self.signatureAlgorithm)
        bb += ", OID = "
        bb += HashAlgorithmConverter.getString(self.signatureAlgorithm)
        bb += os.linesep
        bb += "Issuer: " + self.issuer
        bb += os.linesep
        bb += "Validity: [From: " + str(self.validFrom)
        bb += " GMT To: " + str(self.validTo) + " GMT]" + os.linesep
        bb += "Subject: "
        bb += self.subject
        bb += os.linesep
        bb += "Public Key Algorithm: "
        bb += str(self.publicKeyAlgorithm)
        bb += os.linesep
        bb += "Key: "
        bb += self.publicKey.toHex()
        bb += os.linesep
        if self.publicKey.scheme == Ecc.P256:
            bb += "ASN1 OID: prime256v1" + os.linesep
            bb += "NIST CURVE: P-256"
        elif self.publicKey.scheme == Ecc.P384:
            bb += "ASN1 OID: prime384v1" + os.linesep
            bb += os.linesep
            bb += "NIST CURVE: P-384"
        bb += os.linesep
        bb += "Basic constraints: "
        bb += self.basicConstraints
        bb += os.linesep
        bb += "SubjectKeyIdentifier: "
        bb += GXByteBuffer.hex(self.subjectKeyIdentifier, True)
        bb += os.linesep
        bb += "KeyUsage: "
        bb += str(self.keyUsage)
        bb += os.linesep
        bb += "Signature Algorithm: "
        bb += str(self.signatureAlgorithm)
        bb += os.linesep
        bb += "Signature: "
        bb += GXByteBuffer.hex(self.signature, False)
        bb += os.linesep
        return bb

    @classmethod
    def load(cls, path):
        """
        Load x509 certificate from the PEM file.

            Parameters:
                path: File path.

            Returns:
                Created GXx509Certificate object.
        """
        with open(path, "r", encoding="utf-8") as file:
            return cls.fromPem(file.read())

    def save(self, path):
        """
        x509 certificate to PEM file.

            Parameters:
                path: File path.
        """
        with open(path, "w", encoding="utf-8") as file:
            file.write(self.toPem())

    def toPem(self):
        """
           x509 certificate in PEM format.
        *
        * @return Public key as in PEM string.
        """
        if not self.publicKey:
            raise ValueError("Public or private key is not set.")
        sb = ""
        if self.description:
            sb = "#Description"
            sb += self.description + "\r"
        sb += "-----BEGIN CERTIFICATE-----\r"
        sb += self.toDer()
        sb += "\r-----END CERTIFICATE-----\r"
        return sb

    def toDer(self):
        """
           x509 certificate in DER format.
        *
        * @return Public key as in PEM string.
        """
        return _GXCommon.toBase64(self.encoded)

    def __eq__(self, other):
        return (
            isinstance(other, GXx509Certificate)
            and self.serialNumber == other.serialNumber
        )

    # pylint: disable=import-outside-toplevel,protected-access
    def isCertified(self, certifier):
        """
        Test is x509 file certified by the certifier.

            Parameters:
                certifier: Public key of the certifier.

            Returns:
                True, if certifier has certified the certificate.
        """
        from .GXAsn1Converter import GXAsn1Converter

        if not certifier:
            raise ValueError("certifier")
        # Get raw data
        tmp2 = GXByteBuffer()
        tmp2.set(self.encoded)
        GXAsn1Converter._getNext(tmp2)
        tmp2.size = tmp2.position
        tmp2.position = 1
        _GXCommon.getObjectCount(tmp2)
        e = GXEcdsa(certifier)
        tmp3 = GXAsn1Converter.fromByteArray(self.signature)
        bb = GXByteBuffer()
        if self.signatureAlgorithm == HashAlgorithm.SHA256WITH_ECDSA:
            size = 32
        else:
            size = 48
        # Some implementations might add extra byte. It must removed.
        if len(tmp3[0]).value == size:
            start = 0
        else:
            start = 1
        bb.set(tmp3[0].value, start, size)
        if len(tmp3[1]).value == size:
            start = 0
        else:
            start = 1
        bb.set(tmp3[1].value, start, size)
        return e.verify(bb.array(), tmp2.subArray(tmp2.position, tmp2.available))

    # pylint: disable=import-outside-toplevel, bare-except
    @classmethod
    def search(cls, folder, type_, systemtitle):
        """
        Search x509 Certificate from the PEM file in given folder.

            Parameters:
                folder: Folder to search.
                type: Certificate type.

            Returns:
                Created GXPkcs8 object.
        """
        from .GXAsn1Converter import GXAsn1Converter

        if type_ == CertificateType.DIGITAL_SIGNATURE:
            usage = KeyUsage.DIGITAL_SIGNATURE
        elif type_ == CertificateType.KEY_AGREEMENT:
            usage = KeyUsage.KEY_AGREEMENT
        elif type_ == CertificateType.TLS:
            usage = KeyUsage.DIGITAL_SIGNATURE | KeyUsage.KEY_AGREEMENT
        else:
            usage = KeyUsage.NONE
        subject = GXAsn1Converter.systemTitleToSubject(systemtitle)
        certificates = []
        files = [
            f for f in os.listdir(folder) if os.path.isfile(os.path.join(folder, f))
        ]
        for it in files:
            if it.endswith(".pem") or it.endswith(".cer"):
                try:
                    cert = cls.load(it)
                    if (
                        usage in (KeyUsage.NONE, cert.keyUsage)
                    ) and cert.subject.contains(subject):
                        certificates.append(cert)
                except:
                    pass
        return certificates

    # pylint: disable=import-outside-toplevel
    def getSystemTitle(self):
        """
        Returns system title from the certificate.
        """
        from .GXAsn1Converter import GXAsn1Converter

        if not self.subject:
            return None
        return GXByteBuffer.hexToBytes(
            GXAsn1Converter.hexSystemTitleFromSubject(self.subject)
        )



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .InactivityMode import InactivityMode
from .StartProtocolType import StartProtocolType
from .GXObisCodeCollection import GXObisCodeCollection
#
class GXManufacturer:
    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        """
        Constructor.
        """
        self.inactivityMode = InactivityMode.KEEPALIVE
        self.useIEC47 = False
        self.forceInactivity = False
        self.useLogicalNameReferencing = False
        self.identification = None
        self.obisCodes = GXObisCodeCollection()
        self.name = None
        self.settings = []
        self.serverSettings = []
        self.keepAliveInterval = 40000
        self.startProtocol = StartProtocolType.IEC
        self.webAddress = None
        self.info = None


    def getServer(self, address):
        for it in self.serverSettings:
            if it.HDLCAddress == address:
                return it
        return None

    #
    # Get authentication settings.
    #
    # @param authentication
    #            Authentication type.
    # Authentication settings.
    #
    def getAuthentication(self, authentication):
        for it in self.settings:
            if it.type_ == authentication:
                return it
        return None



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import datetime
from gurux_dlms.GXByteBuffer import GXByteBuffer
from gurux_dlms.objects.enums.CertificateType import CertificateType
from gurux_dlms.internal._GXCommon import _GXCommon
from gurux_dlms.enums.BerType import BerType
from .GXAsn1Context import GXAsn1Context
from .enums.X509Name import X509Name
from .GXAsn1Ia5String import GXAsn1Ia5String
from .GXAsn1Utf8String import GXAsn1Utf8String
from .GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier
from .X509NameConverter import X509NameConverter
from .GXAsn1Sequence import GXAsn1Sequence
from .GXAsn1Integer import GXAsn1Integer
from ..GXBitString import GXBitString
from .enums.KeyUsage import KeyUsage
from .GXx509Certificate import GXx509Certificate
from .enums.PkcsType import PkcsType
from ..GXInt8 import GXInt8
from ..GXInt16 import GXInt16
from ..GXInt32 import GXInt32
from ..GXInt64 import GXInt64
from ..GXArray import GXArray
from ..asn.GXAsn1PublicKey import GXAsn1PublicKey


class GXAsn1Converter:
    """
    ASN1 converter. This class is used to convert public and private keys to byte array and vice verse.
    """

    @classmethod
    def encodeSubject(cls, value):
        list_ = []
        for tmp in value.split(","):
            it = tmp.split("=")
            if len(it) != 2:
                raise ValueError("Invalid subject.")
            name = X509Name.valueofString(it[0].strip())
            if name == X509Name.C:
                # Country code is printable string
                val = it[1]
            elif name == X509Name.E:
                # email address in Verisign certificates
                val = GXAsn1Ia5String(it[1].strip())
            else:
                val = GXAsn1Utf8String(it[1].strip())
            oid = X509NameConverter.getString(name)
            list_.append((GXAsn1ObjectIdentifier(oid), val))
        return list_

    @classmethod
    def getSubject(cls, values):
        sb = ""
        first = True
        for k, v in values:
            if first:
                first = False
            else:
                sb += ", "
            sb += str(X509NameConverter.fromString(str(k)))
            sb += "="
            sb += str(v)
        return sb

    # pylint: disable=too-many-locals
    @classmethod
    def __getValue(cls, bb, objects, s, getNext):
        type_ = bb.getUInt8()
        len_ = _GXCommon.getObjectCount(bb)
        if len_ > bb.available():
            raise ValueError("GXAsn1Converter.GetValue")
        connectPos = 0
        if s:
            connectPos = s.xmlLength
        start = bb.position
        tagString = ""
        if s:
            s.appendSpaces()
            if type == BerType.INTEGER:
                if len_ in (1, 2, 4, 8):
                    tagString = s.getTag(-len_)
                else:
                    tagString = s.getTag(BerType.INTEGER)
            else:
                tagString = s.GetTag(type_)
            s.append("<" + tagString + ">")

        if type_ in (
            (BerType.CONSTRUCTED | BerType.CONTEXT),
            (BerType.CONSTRUCTED | BerType.CONTEXT | 1),
            (BerType.CONSTRUCTED | BerType.CONTEXT | 2),
            (BerType.CONSTRUCTED | BerType.CONTEXT | 3),
            (BerType.CONSTRUCTED | BerType.CONTEXT | 4),
            (BerType.CONSTRUCTED | BerType.CONTEXT | 5),
        ):
            if s:
                s.increase()
            tmp = GXAsn1Context()
            tmp.index = type_ & 0xF
            objects.append(tmp)
            while bb.position < start + len_:
                cls.__getValue(bb, tmp, s, False)
            if s:
                s.decrease()
        elif type_ == (BerType.CONSTRUCTED | BerType.SEQUENCE):
            if s:
                s.increase()
            tmp = GXAsn1Sequence()
            objects.append(tmp)
            cnt = 0
            while bb.position < start + len_:
                cnt += 1
                cls.__getValue(bb, tmp, s, False)
                if getNext:
                    break
            if s:
                # Append comment.
                s.appendComment(connectPos, str(cnt) + " elements.")
                s.decrease()
        elif type_ == (BerType.CONSTRUCTED | BerType.SET):
            if s:
                s.increase()
            tmp = []
            cls.__getValue(bb, tmp, s, False)
            if isinstance(tmp[0], GXAsn1Sequence):
                tmp = tmp[0]
                v = None
                if len(tmp) != 1:
                    v = tmp[1]
                objects.append((tmp[0], v))
            else:
                e = (tmp, None)
                objects.append(e)
            if s:
                s.decrease()
        elif type_ == BerType.OBJECT_IDENTIFIER:
            oi = GXAsn1ObjectIdentifier(bb, len_)
            objects.append(oi)
            if s:
                if oi.description:
                    s.appendComment(connectPos, oi.description)
                s.append(str(oi))
        elif type_ == BerType.PRINTABLE_STRING:
            str_ = bb.getString(bb.position, len_)
            bb.position += len_
            objects.append(str_)
            if s:
                s.Append(str_)
        elif type_ == BerType.BMP_STRING:
            str_ = bb.GetStringUnicode(len_)
            objects.append(str_)
            if s:
                s.Append(str_)
        elif type_ == BerType.UTF8STRING:
            objects.append(GXAsn1Utf8String(bb.getString(bb.position, len_)))
            bb.position += len_
            if s:
                s.append(str(objects[len(objects) - 1]))
        elif type_ == BerType.IA5_STRING:
            objects.append(GXAsn1Ia5String(bb.getString(len_)))
            if s:
                s.Append(str(objects[len(objects) - 1]))
        elif type_ == BerType.INTEGER:
            if len_ == 1:
                objects.append(bb.getInt8())
            elif len_ == 2:
                objects.append(bb.getInt16())
            elif len_ == 4:
                objects.append(bb.getInt32())
            else:
                tmp2 = bytearray(len_)
                bb.get(tmp2)
                objects.append(GXAsn1Integer(tmp2))
            if s:
                s.Append(str(objects[len(objects) - 1]))
        elif type_ == BerType.NULL:
            objects.append(None)
        elif type_ == BerType.BIT_STRING:
            tmp3 = GXBitString(bb.subArray(bb.position, len_))
            objects.append(tmp3)
            bb.position += len_
            if s:
                # Append comment.
                s.appendComment(connectPos, str(len(tmp3)) + " bit.")
                s.append(str(tmp3))
        elif type_ == BerType.UTC_TIME:
            tmp2 = bytearray(len_)
            bb.get(tmp2)
            objects.append(cls.__getUtcTime(tmp2.decode("utf-8")))
            if s:
                s.append(str(objects[objects.Count - 1]))
        elif type_ == BerType.GENERALIZED_TIME:
            tmp2 = bytes(len_)
            bb.get(tmp2)
            objects.append(_GXCommon.getGeneralizedTime(str(tmp2)))
            if s:
                s.append(str(objects[objects.Count - 1]))
        elif type_ in (
            BerType.CONTEXT,
            (BerType.CONTEXT | 1),
            (BerType.CONTEXT | 2),
            (BerType.CONTEXT | 3),
            (BerType.CONTEXT | 4),
            (BerType.CONTEXT | 5),
            (BerType.CONTEXT | 6),
        ):
            tmp = GXAsn1Context()
            tmp.constructed = False
            tmp.index = type_ & 0xF
            tmp2 = bytearray(len_)
            bb.get(tmp2)
            tmp.append(tmp2)
            objects.append(tmp)
            if s:
                s.append(GXByteBuffer.toHex(tmp2, False))
        elif type_ == BerType.OCTET_STRING:
            t = bb.getUInt8(bb.position)
            if t in ((BerType.CONSTRUCTED | BerType.SEQUENCE), BerType.BIT_STRING):
                if s:
                    s.increase()
                cls.__getValue(bb, objects, s, False)
                if s:
                    s.decrease()
            else:
                tmp2 = bytearray(len_)
                bb.get(tmp2)
                objects.append(tmp2)
                if s:
                    s.append(GXByteBuffer.toHex(tmp2, False))
        elif type_ == BerType.BOOLEAN:
            b = bb.getUInt8() != 0
            objects.append(b)
            if s:
                s.append(str(b))
        else:
            raise ValueError("Invalid type: " + str(type_))
        if s:
            s.append("</" + tagString + ">\r\n")

    @classmethod
    def __getUtcTime(cls, dateString):
        year = 2000 + int(dateString[0:2])
        month = int(dateString[2:4])
        day = int(dateString[4:6])
        hour = int(dateString[6:8])
        minute = int(dateString[8:10])
        # If UTC time.
        if dateString.find("Z") != -1:
            second = 0
            if len(dateString) > 11:
                second = int(dateString[10:12])
            return datetime.datetime(year, month, day, hour, minute, second, 0)
        if len(dateString) > 15:
            second = int(dateString[10:12])
        return datetime.datetime(year, month, day, hour, minute, second, 0)

    @classmethod
    def __dateToString(cls, date):
        sb = f"{date.year - 2000:02d}"
        sb += f"{date.month:02d}"
        sb += f"{date.day:02d}"
        sb += f"{date.hour:02d}"
        sb += f"{date.minute:02d}"
        sb += f"{date.second:02d}"
        sb += "Z"
        return sb

    @classmethod
    def fromByteArray(cls, data):
        """
        Convert byte array to ASN1 objects.

        Parameters:
            data: ASN-1 bytes.

        Returns:
            Parsed objects.
        """
        bb = GXByteBuffer(data)
        objects = []
        while bb.position != bb.size:
            cls.__getValue(bb, objects, None, False)

        if not objects:
            return None
        return objects[0]

    # Get next ASN1 value from the byte buffer.
    @classmethod
    def _getNext(cls, data):
        objects = []
        cls.__getValue(data, objects, None, True)
        return objects[0]

    # add ASN1 object to byte buffer.
    @classmethod
    def __getBytes(cls, bb, target):
        start = bb.size
        cnt = 0
        if isinstance(target, GXAsn1Context):
            tmp = GXByteBuffer()
            for it in target:
                cnt += cls.__getBytes(tmp, it)
            start = bb.size
            if target.constructed:
                bb.setUInt8(BerType.CONSTRUCTED | BerType.CONTEXT | target.index)
                _GXCommon.setObjectCount(cnt, bb)
            else:
                tmp.setUInt8(0, BerType.CONTEXT | target.index)
            cnt += bb.size - start
            bb.set(tmp)
            return cnt
        if isinstance(target, GXArray):
            tmp = GXByteBuffer()
            for it in target:
                cnt += cls.__getBytes(tmp, it)
            start = bb.size
            bb.setUInt8(BerType.CONSTRUCTED | BerType.SEQUENCE)
            _GXCommon.setObjectCount(cnt, bb)
            cnt += bb.size - start
            bb.set(tmp)
            return cnt
        if isinstance(target, str):
            bb.setUInt8(BerType.PRINTABLE_STRING)
            _GXCommon.setObjectCount(len(target), bb)
            bb.set(target)
        elif isinstance(target, GXInt8):
            bb.setUInt8(BerType.INTEGER)
            _GXCommon.setObjectCount(1, bb)
            bb.setInt8(target)
        elif isinstance(target, GXInt16):
            bb.setUInt8(BerType.INTEGER)
            _GXCommon.setObjectCount(2, bb)
            bb.setInt16(target)
        elif isinstance(target, GXInt32):
            bb.setUInt8(BerType.INTEGER)
            _GXCommon.setObjectCount(4, bb)
            bb.setInt32(target)
        elif isinstance(target, GXAsn1Integer):
            bb.setUInt8(BerType.INTEGER)
            b = target.toArray()
            _GXCommon.setObjectCount(len(b), bb)
            bb.set(b)
        elif isinstance(target, GXInt64):
            bb.setUInt8(BerType.INTEGER)
            _GXCommon.setObjectCount(8, bb)
            bb.setInt64(target)
        elif isinstance(target, (bytes, bytearray)):
            bb.setUInt8(BerType.OCTET_STRING)
            _GXCommon.setObjectCount(len(target), bb)
            bb.set(target)
        elif target is None:
            bb.setUInt8(BerType.NULL)
            _GXCommon.setObjectCount(0, bb)
        elif isinstance(target, bool):
            bb.setUInt8(BerType.BOOLEAN)
            bb.setUInt8(1)
            if target:
                bb.setUInt8(255)
            else:
                bb.setUInt8(0)
        elif isinstance(target, GXAsn1ObjectIdentifier):
            bb.setUInt8(BerType.OBJECT_IDENTIFIER)
            t = target.encoded
            _GXCommon.setObjectCount(len(t), bb)
            bb.set(t)
        elif isinstance(target, tuple):
            tmp2 = GXByteBuffer()
            if target[1]:
                tmp = GXByteBuffer()
                cnt += cls.__getBytes(tmp2, target[0])
                cnt += cls.__getBytes(tmp2, target[1])
                tmp.setUInt8(BerType.CONSTRUCTED | BerType.SEQUENCE)
                _GXCommon.setObjectCount(cnt, tmp)
                tmp.set(tmp2)
            else:
                cls.__getBytes(tmp2, target[0])
                tmp = tmp2
            # Update len.
            cnt = bb.size
            bb.setUInt8(BerType.CONSTRUCTED | BerType.SET)
            _GXCommon.setObjectCount(tmp.size, bb)
            bb.set(tmp)
            return bb.size - cnt
        elif isinstance(target, GXAsn1Utf8String):
            bb.setUInt8(BerType.UTF8STRING)
            str_ = str(target)
            _GXCommon.setObjectCount(len(str_), bb)
            bb.set(str_)
        elif isinstance(target, GXAsn1Ia5String):
            bb.setUInt8(BerType.IA5_STRING)
            _GXCommon.setObjectCount(len(target), bb)
            bb.set(target)
        elif isinstance(target, GXBitString):
            bb.setUInt8(BerType.BIT_STRING)
            _GXCommon.setObjectCount(1 + len(target.value), bb)
            bb.setUInt8(target.padBits)
            bb.set(target.value)
        elif isinstance(target, GXAsn1PublicKey):
            bb.setUInt8(BerType.BIT_STRING)
            # size is 64 bytes + padding and uncompressed point indicator.
            _GXCommon.setObjectCount(66, bb)
            # add padding.
            bb.setUInt8(0)
            # prefixed with the uncompressed point indicator 04
            bb.setUInt8(4)
            bb.add(target.value)
            # Count is type + size + 64 bytes + padding + uncompressed point
            # indicator.
            return 68
        elif isinstance(target, datetime.datetime):
            # Save date time in UTC.
            bb.setUInt8(BerType.UTC_TIME)
            str_ = cls.__dateToString(target)
            bb.setUInt8(len(str_))
            bb.add(str_)
        elif isinstance(target, (GXAsn1Sequence, list)):
            tmp = GXByteBuffer()
            for it in target:
                cnt += cls.__getBytes(tmp, it)
            start = bb.size
            if isinstance(target, GXAsn1Context):
                if target.constructed:
                    bb.setUInt8((BerType.CONSTRUCTED | BerType.SEQUENCE | target.Index))
                else:
                    bb.setUInt8(BerType.SEQUENCE | target.index)
            else:
                bb.setUInt8(BerType.CONSTRUCTED | BerType.SEQUENCE)
            _GXCommon.setObjectCount(cnt, bb)
            cnt += bb.size - start
            bb.set(tmp)
            return cnt
        else:
            raise ValueError("Invalid type: " + str(target))
        return bb.size - start

    # Convert ASN1 objects to byte array.
    @classmethod
    def toByteArray(cls, objects):
        bb = GXByteBuffer()
        cls.__getBytes(bb, objects)
        return bb.array()

    @classmethod
    def systemTitleToSubject(cls, systemTitle):
        """
        Convert system title to subject.

            Parameters:
                systemTitle: System title.

            Returns:
                Subject.
        """
        bb = GXByteBuffer(systemTitle)
        return "CN=" + bb.toHex(False, 0)

    @classmethod
    def systemTitleFromSubject(cls, subject):
        """
        Get system title from the subject.

            Parameters:
                subject: Subject.

            Returns:
                System title.
        """
        return GXByteBuffer.hexToBytes(cls.hexSystemTitleFromSubject(subject))

    @classmethod
    def hexSystemTitleFromSubject(cls, subject):
        """
        Get system title in hex string from the subject.

            Parameters:
                subject: Subject.

            Returns:
                System title.
        """
        index = subject.find("CN=")
        if index == -1:
            raise ValueError("System title not found from the subject.")
        return subject[index + 3 : index + 3 + 16]

    @classmethod
    def certificateTypeToKeyUsage(cls, value):
        """
        Convert ASN1 certificate type to DLMS key usage.

            Parameters:
                type: Certificate type.

            Returns:
                Key usage.
        """
        if value == CertificateType.DIGITAL_SIGNATURE:
            k = KeyUsage.DIGITAL_SIGNATURE
        elif value == CertificateType.KEY_AGREEMENT:
            k = KeyUsage.KEY_AGREEMENT
        elif value == CertificateType.TLS:
            k = KeyUsage.KEY_CERT_SIGN
        elif value == CertificateType.OTHER:
            k = KeyUsage.CRL_SIGN
        else:
            # At least one bit must be used.
            k = KeyUsage.NONE
        return k

    @classmethod
    def getCertificateType(cls, value):
        """
        Get certificate type from byte array or DER string.

        Parameters:
            value: Byte array or DER string.

        Returns:
            Certificate type.
        """
        if isinstance(value, str):
            return cls.__getCertificateType(_GXCommon.fromBase64(value), None)
        return cls.__getCertificateType(value, None)

    # pylint: disable=bare-except, import-outside-toplevel
    @classmethod
    def __getCertificateType(cls, data, seq):
        if not seq:
            seq = GXAsn1Converter.fromByteArray(data)
        if isinstance(seq[0], GXAsn1Sequence):
            try:
                GXx509Certificate(data)
                return PkcsType.X509CERTIFICATE
            except:
                # It's ok if this fails.
                pass
        if isinstance(seq[0], GXAsn1Sequence):
            try:
                from .GXPkcs10 import GXPkcs10

                GXPkcs10(data)
                return PkcsType.PKCS10
            except:
                # It's ok if this fails.
                pass
        if isinstance(seq[0], int):
            try:
                from .GXPkcs8 import GXPkcs8

                GXPkcs8(data)
                return PkcsType.PKCS8
            except:
                # It's ok if this fails.
                pass
        return PkcsType.NONE



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
class GXDLMSMacPosTable:
    __shortAddress = None
    "The 16-bit address the device is using to communicate through the PAN."

    __lQI = 0
    "Link Quality Indicator."

    __validTime = None
    "Valid time."

    @property
    def shortAddress(self):
        """
        The 16-bit address the device is using to communicate through the PAN.
        """
        return self.__shortAddress

    @shortAddress.setter
    def shortAddress(self, value):
        self.__shortAddress = value

    @property
    def lQI(self):
        """
        Link Quality Indicator.
        """
        return self.__lQI

    @lQI.setter
    def lQI(self, value):
        self.__lQI = value

    @property
    def validTime(self):
        """
        Valid time,
        """
        return self.__validTime

    @validTime.setter
    def validTime(self, value):
        self.__validTime = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------

from gurux_dlms.enums.BerType import BerType


class GXAsn1Settings:
    __count = 0

    __comments = None
    """
    Are comments used.
    """

    __sb = ""

    __tags = {}

    __tagbyName = {}

    @property
    def comments(self):
        """
        Are comments used.
        """
        return self.__comments

    @comments.setter
    def comments(self, value):
        self.__comments = value

    @property
    def xmlLength(self):
        return len(self.__sb)

    def __init__(self):
        """
        Constructor.
        """

        self.__addTag(BerType.APPLICATION, "Application")
        self.__addTag((BerType.CONSTRUCTED | BerType.CONTEXT), "Context")
        self.__addTag(
            ((BerType.CONSTRUCTED | BerType.CONTEXT | 1)),
            "Constructed-Context-Specific-1",
        )
        self.__addTag(
            ((BerType.CONSTRUCTED | BerType.CONTEXT | 2)),
            "Constructed-Context-Specific-2",
        )
        self.__addTag(
            ((BerType.CONSTRUCTED | BerType.CONTEXT | 3)),
            "Constructed-Context-Specific-3",
        )
        self.__addTag((BerType.CONSTRUCTED | BerType.SEQUENCE), "Sequence")
        self.__addTag((BerType.CONSTRUCTED | BerType.SET), "Set")
        self.__addTag(BerType.OBJECT_IDENTIFIER, "ObjectIdentifier")
        self.__addTag(BerType.PRINTABLE_STRING, "String")
        self.__addTag(BerType.UTF8STRING, "UTF8")
        self.__addTag(BerType.IA5_STRING, "IA5")
        self.__addTag(BerType.INTEGER, "Integer")
        self.__addTag(BerType.NULL, "Null")
        self.__addTag(BerType.BIT_STRING, "BitString")
        self.__addTag(BerType.UTC_TIME, "UtcTime")
        self.__addTag(BerType.GENERALIZED_TIME, "GeneralizedTime")
        self.__addTag(BerType.OCTET_STRING, "OctetString")
        self.__addTag(BerType.BOOLEAN, "Bool")
        self.__addTag(-1, "Byte")
        self.__addTag(-2, "Short")
        self.__addTag(-4, "Int")
        self.__addTag(-8, "Long")

    def __addTag(self, key, value):
        self.__tags[key] = value
        self.__tagbyName[value.lower()] = key

    def getTag(self, value):
        if isinstance(value, str):
            return self.__tagbyName[value]
        return self.__tags[value]

    # pylint: disable=unused-variable
    def appendComment(self, offset, value):
        """
        Add comment.

            Parameters:
                offset: Offset.
                value: Comment value.
        """
        if self.comments:
            empty_ = len(self.__sb) == 0
            if empty_:
                tmp = self.__sb
            else:
                tmp = ""

            for pos in range(0, self.__count - 1):
                tmp += " "
            tmp += "<!--"
            tmp += value
            tmp += "-->\r\n"
            if not empty_:
                self.__sb = self.__sb[:offset] + tmp + self.__sb[offset:]

    # pylint: disable=unused-variable
    def appendSpaces(self):
        """
        Append spaces to the buffer.
        """
        for pos in range(0, self.__count - 1):
            __sb += " "

    def append(self, value):
        self.__sb += value

    def increase(self):
        self.__count += 1
        self.append("\r\n")

    def decrease(self):
        self.__count -= 1
        self.appendSpaces()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .GXDLMSMonitoredValue import GXDLMSMonitoredValue
from .GXDLMSActionSet import GXDLMSActionSet
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSRegisterMonitor(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSRegisterMonitor
    """

    def __init__(self, ln=None, sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        # pylint: disable=super-with-arguments
        super(GXDLMSRegisterMonitor, self).__init__(ObjectType.REGISTER_MONITOR, ln, sn)
        self.thresholds = []
        self.monitoredValue = GXDLMSMonitoredValue()
        self.actions = []

    def getValues(self):
        return (self.logicalName, self.thresholds, self.monitoredValue, self.actions)

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Thresholds
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  MonitoredValue
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  Actions
        if all_ or not self.isRead(4):
            attributes.append(4)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 4

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Thresholds"),\
            _GXLocalizer.gettext("Monitored value"),\
            _GXLocalizer.gettext("Actions"))

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index == 2:
            return super(GXDLMSRegisterMonitor, self).getDataType(index)
        if index == 3:
            return DataType.ARRAY
        if index == 4:
            return DataType.ARRAY
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.thresholds
        if e.index == 3:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(3)
            #  ClassID
            _GXCommon.setData(
                settings, bb, DataType.UINT16, self.monitoredValue.objectType
            )
            #  LN.
            _GXCommon.setData(
                settings,
                bb,
                DataType.OCTET_STRING,
                _GXCommon.logicalNameToBytes(self.monitoredValue.logicalName),
            )
            #  Attribute index.
            _GXCommon.setData(
                settings, bb, DataType.INT8, self.monitoredValue.attributeIndex
            )
            return bb
        if e.index == 4:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            if not self.actions:
                bb.setUInt8(0)
            else:
                bb.setUInt8(len(self.actions))
                for it in self.actions:
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(2)
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(2)
                    #  LN
                    _GXCommon.setData(
                        settings,
                        bb,
                        DataType.OCTET_STRING,
                        _GXCommon.logicalNameToBytes(it.actionUp.logicalName),
                    )
                    #  ScriptSelector
                    _GXCommon.setData(
                        settings, bb, DataType.UINT16, it.actionUp.scriptSelector
                    )
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(2)
                    #  LN
                    _GXCommon.setData(
                        settings,
                        bb,
                        DataType.OCTET_STRING,
                        _GXCommon.logicalNameToBytes(it.actionDown.logicalName),
                    )
                    #  ScriptSelector
                    _GXCommon.setData(
                        settings, bb, DataType.UINT16, it.actionDown.scriptSelector
                    )
            return bb
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.thresholds = e.value
        elif e.index == 3:
            if self.monitoredValue is None:
                self.monitoredValue = GXDLMSMonitoredValue()
            self.monitoredValue.objectType = e.value[0]
            self.monitoredValue.logicalName = _GXCommon.toLogicalName(e.value[1])
            self.monitoredValue.attributeIndex = e.value[2]
        elif e.index == 4:
            self.actions = []
            if e.value:
                for as_ in e.value:
                    set_ = GXDLMSActionSet()
                    target = as_[0]
                    set_.actionUp.logicalName = _GXCommon.toLogicalName(target[0])
                    set_.actionUp.scriptSelector = target[1]
                    target = as_[1]
                    set_.actionDown.logicalName = _GXCommon.toLogicalName(target[0])
                    set_.actionDown.scriptSelector = target[1]
                    self.actions.append(set_)
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.thresholds = []
        if reader.isStartElement("Thresholds", True):
            while reader.isStartElement("Value", False):
                it = reader.readElementContentAsObject("Value", None)
                self.thresholds.append(it)
            reader.readEndElement("Thresholds")
        if reader.isStartElement("MonitoredValue", True):
            self.monitoredValue.objectType = ObjectType(
                reader.readElementContentAsInt("ObjectType")
            )
            self.monitoredValue.logicalName = reader.readElementContentAsString("LN")
            self.monitoredValue.attributeIndex = reader.readElementContentAsInt("Index")
            reader.readEndElement("MonitoredValue")
        self.actions = []
        if reader.isStartElement("Actions", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSActionSet()
                self.actions.append(it)
                if reader.isStartElement("Up", True):
                    it.actionUp.logicalName = reader.readElementContentAsString(
                        "LN", None
                    )
                    it.actionUp.scriptSelector = reader.readElementContentAsInt(
                        "Selector"
                    )
                    reader.readEndElement("Up")
                if reader.isStartElement("Down", True):
                    it.actionDown.logicalName = reader.readElementContentAsString(
                        "LN", None
                    )
                    it.actionDown.scriptSelector = reader.readElementContentAsInt(
                        "Selector"
                    )
                    reader.readEndElement("Down")
            reader.readEndElement("Actions")

    def save(self, writer):
        if self.thresholds:
            writer.writeStartElement("Thresholds")
            for it in self.thresholds:
                writer.writeElementObject("Value", it)
            writer.writeEndElement()
        if self.monitoredValue:
            writer.writeStartElement("MonitoredValue")
            writer.writeElementString("ObjectType", int(self.monitoredValue.objectType))
            writer.writeElementString("LN", self.monitoredValue.logicalName)
            writer.writeElementString("Index", self.monitoredValue.attributeIndex)
            writer.writeEndElement()
        if self.actions:
            writer.writeStartElement("Actions")
            for it in self.actions:
                writer.writeStartElement("Item")
                writer.writeStartElement("Up")
                writer.writeElementString("LN", it.actionUp.logicalName)
                writer.writeElementString("Selector", it.actionUp.scriptSelector)
                writer.writeEndElement()
                writer.writeStartElement("Down")
                writer.writeElementString("LN", it.actionDown.logicalName)
                writer.writeElementString("Selector", it.actionDown.scriptSelector)
                writer.writeEndElement()
                writer.writeEndElement()
            writer.writeEndElement()



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {'__slots__': ()})

class IGXDLMSBase(ABC):

    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    # @param all_
    #            All are attributes get.
    # @return Collection of attributes to read.
    @abstractmethod
    def getAttributeIndexToRead(self, all_):
        """ This is interface method """

    # @return Amount of attributes.
    @abstractmethod
    def getAttributeCount(self):
        """ This is interface method """

    # @return Amount of methods.
    @abstractmethod
    def getMethodCount(self):
        """ Amount of methods. """

    # @return Names of attribute indexes.
    @abstractmethod
    def getNames(self):
        """ Names of attribute indexes. """

    # @return Names of attribute indexes.
    @abstractmethod
    def getMethodNames(self):
        """ Names of attribute indexes """

    # Returns value of given attribute.
    #
    # @param settings
    #            DLMS settings.
    # @param e
    #            Get parameter.
    # @return Returned value.
    @abstractmethod
    def getValue(self, settings, e):
        """ This is interface method """

    # Set value of given attribute.
    #
    # @param settings
    #            DLMS settings.
    # @param e
    #            Set parameter.
    @abstractmethod
    def setValue(self, settings, e):
        """ This is interface method """

    # Server calls invoke method.
    #
    # @param settings
    #            Server settings.
    # @param e
    #            Invoke parameter.
    # @return Reply for the client.
    @abstractmethod
    def invoke(self, settings, e):
        """ This is interface method """

    # Load object content from XML.
    #
    # @param reader
    #            XML reader.
    @abstractmethod
    def load(self, reader):
        """ This is interface method """

    # Save object content to XML.
    #
    # @param writer
    #            XML writer.
    @abstractmethod
    def save(self, writer):
        """ This is interface method """

    # Handle actions after Load.
    #
    # @param reader
    #            XML reader.
    def postLoad(self, reader):
        """ This is interface method """



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.GXIntEnum import GXIntEnum


class X9ObjectIdentifier(GXIntEnum):
    NONE = 0
    ID_FIELD_TYPE = 1
    PRIME_FIELD = 2
    CHARACTERISTIC_TWO_FIELD = 3
    GN_BASIS = 4
    TP_BASIS = 5
    PP_BASIS = 6
    EC_DSA_WITH_SHA1 = 7
    ID_EC_PUBLIC_KEY = 8
    EC_DSA_WITH_SHA2 = 9
    EC_DSA_WITH_SHA224 = 10
    EC_DSA_WITH_SHA256 = 11
    EC_DSA_WITH_SHA384 = 12
    EC_DSA_WITH_SHA512 = 13
    ELLIPTIC_CURVE = 14
    C_TWO_CURVE = 15
    C2PNB163V1 = 16
    C2PNB163V2 = 17
    C2PNB163V3 = 18
    C2PNB176W1 = 19
    C2TNB191V1 = 20
    C2TNB191V2 = 21
    C2TNB191V3 = 22
    C2ONB191V4 = 23
    C2ONB191V5 = 24
    C2PNB208W1 = 25
    C2TNB239V1 = 26
    C2TNB239V2 = 27
    C2TNB239V3 = 28
    C2ONB239V4 = 29
    C2ONB239V5 = 30
    C2PNB272W1 = 31
    C2PNB304W1 = 32
    C2TNB359V1 = 33
    C2PNB368W1 = 34
    C2TNB431R1 = 35
    PRIME_CURVE = 36
    PRIME192V1 = 37
    PRIME192V2 = 38
    PRIME192V3 = 39
    PRIME239V1 = 40
    PRIME239V2 = 41
    PRIME239V3 = 42
    PRIME256V1 = 43
    ID_DSA = 44
    ID_DSA_WITH_SHA1 = 45
    X9X63SCHEME = 46
    DH_SINGLE_PASS_STD_DH_SHA1KDF_SCHEME = 47
    DH_SINGLE_PASS_COFACTOR_DH_SHA1KDF_SCHEME = 48
    MQV_SINGLE_PASS_SHA1KDF_SCHEME = 49
    ANSI_X9_42 = 50
    DH_PUBLIC_NUMBER = 51
    X9X42SCHEMES = 52
    DH_STATIC = 53
    DH_EPHEM = 54
    DH_ONE_FLOW = 55
    DH_HYBRID1 = 56
    DH_HYBRID2 = 57
    DH_HYBRID_ONE_FLOW = 58
    MQV2 = 59
    MQV1 = 60
    SECP384R1 = 61

    @classmethod
    def valueofString(cls, value):
        return X9ObjectIdentifier[value.upper()]

    def __str__(self):
        if self.value == X9ObjectIdentifier.NONE:
            ret = "None"
        elif self.value == X9ObjectIdentifier.ID_FIELD_TYPE:
            ret = "Id field type"
        elif self.value == X9ObjectIdentifier.PRIME_FIELD:
            ret = "Prime field"
        elif self.value == X9ObjectIdentifier.CHARACTERISTIC_TWO_FIELD:
            ret = "Characteristic two field"
        elif self.value == X9ObjectIdentifier.GN_BASIS:
            ret = "Gn basis"
        elif self.value == X9ObjectIdentifier.TP_BASIS:
            ret = "Tp basis"
        elif self.value == X9ObjectIdentifier.PP_BASIS:
            ret = "Pp basis"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA1:
            ret = "Ec dsa with sha1"
        elif self.value == X9ObjectIdentifier.ID_EC_PUBLIC_KEY:
            ret = "Id ec public key"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA2:
            ret = "Ec dsa with sha2"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA224:
            ret = "Ec dsa with sha224"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA256:
            ret = "Ec dsa with sha256"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA384:
            ret = "Ec dsa with sha384"
        elif self.value == X9ObjectIdentifier.EC_DSA_WITH_SHA512:
            ret = "Ec dsa with sha512"
        elif self.value == X9ObjectIdentifier.ELLIPTIC_CURVE:
            ret = "Elliptic curve"
        elif self.value == X9ObjectIdentifier.C_TWO_CURVE:
            ret = "C two curve"
        elif self.value == X9ObjectIdentifier.C2PNB163V1:
            ret = "C2pnb163v1"
        elif self.value == X9ObjectIdentifier.C2PNB163V2:
            ret = "C2pnb163v2"
        elif self.value == X9ObjectIdentifier.C2PNB163V3:
            ret = "C2pnb163v3"
        elif self.value == X9ObjectIdentifier.C2PNB176W1:
            ret = "C2pnb176w1"
        elif self.value == X9ObjectIdentifier.C2TNB191V1:
            ret = "C2tnb191v1"
        elif self.value == X9ObjectIdentifier.C2TNB191V2:
            ret = "C2tnb191v2"
        elif self.value == X9ObjectIdentifier.C2TNB191V3:
            ret = "C2tnb191v3"
        elif self.value == X9ObjectIdentifier.C2ONB191V4:
            ret = "C2onb191v4"
        elif self.value == X9ObjectIdentifier.C2ONB191V5:
            ret = "C2onb191v5"
        elif self.value == X9ObjectIdentifier.C2PNB208W1:
            ret = "C2pnb208w1"
        elif self.value == X9ObjectIdentifier.C2TNB239V1:
            ret = "C2tnb239v1"
        elif self.value == X9ObjectIdentifier.C2TNB239V2:
            ret = "C2tnb239v2"
        elif self.value == X9ObjectIdentifier.C2TNB239V3:
            ret = "C2tnb239v3"
        elif self.value == X9ObjectIdentifier.C2ONB239V4:
            ret = "C2onb239v4"
        elif self.value == X9ObjectIdentifier.C2ONB239V5:
            ret = "C2onb239v5"
        elif self.value == X9ObjectIdentifier.C2PNB272W1:
            ret = "C2pnb272w1"
        elif self.value == X9ObjectIdentifier.C2PNB304W1:
            ret = "C2pnb304w1"
        elif self.value == X9ObjectIdentifier.C2TNB359V1:
            ret = "C2tnb359v1"
        elif self.value == X9ObjectIdentifier.C2PNB368W1:
            ret = "C2pnb368w1"
        elif self.value == X9ObjectIdentifier.C2TNB431R1:
            ret = "C2tnb431r1"
        elif self.value == X9ObjectIdentifier.PRIME_CURVE:
            ret = "Prime curve"
        elif self.value == X9ObjectIdentifier.PRIME192V1:
            ret = "Prime192v1"
        elif self.value == X9ObjectIdentifier.PRIME192V2:
            ret = "Prime192v2"
        elif self.value == X9ObjectIdentifier.PRIME192V3:
            ret = "Prime192v3"
        elif self.value == X9ObjectIdentifier.PRIME239V1:
            ret = "Prime239v1"
        elif self.value == X9ObjectIdentifier.PRIME239V2:
            ret = "Prime239v2"
        elif self.value == X9ObjectIdentifier.PRIME239V3:
            ret = "Prime239v3"
        elif self.value == X9ObjectIdentifier.PRIME256V1:
            ret = "Prime256v1"
        elif self.value == X9ObjectIdentifier.ID_DSA:
            ret = "Id dsa"
        elif self.value == X9ObjectIdentifier.ID_DSA_WITH_SHA1:
            ret = "Id dsa with sha1"
        elif self.value == X9ObjectIdentifier.X9X63SCHEME:
            ret = "X9x63scheme"
        elif self.value == X9ObjectIdentifier.DH_SINGLE_PASS_STD_DH_SHA1KDF_SCHEME:
            ret = "Dh single pass std dh sha1kdf scheme"
        elif self.value == X9ObjectIdentifier.DH_SINGLE_PASS_COFACTOR_DH_SHA1KDF_SCHEME:
            ret = "Dh single pass cofactor dh sha1kdf scheme"
        elif self.value == X9ObjectIdentifier.MQV_SINGLE_PASS_SHA1KDF_SCHEME:
            ret = "Mqv single pass sha1kdf scheme"
        elif self.value == X9ObjectIdentifier.ANSI_X9_42:
            ret = "ansi_x9_42"
        elif self.value == X9ObjectIdentifier.DH_PUBLIC_NUMBER:
            ret = "Dh public number"
        elif self.value == X9ObjectIdentifier.X9X42SCHEMES:
            ret = "X9x42schemes"
        elif self.value == X9ObjectIdentifier.DH_STATIC:
            ret = "Dh static"
        elif self.value == X9ObjectIdentifier.DH_EPHEM:
            ret = "Dh ephem"
        elif self.value == X9ObjectIdentifier.DH_ONE_FLOW:
            ret = "Dh one flow"
        elif self.value == X9ObjectIdentifier.DH_HYBRID1:
            ret = "Dh hybrid1"
        elif self.value == X9ObjectIdentifier.DH_HYBRID2:
            ret = "Dh hybrid2"
        elif self.value == X9ObjectIdentifier.DH_HYBRID_ONE_FLOW:
            ret = "Dh hybrid one flow"
        elif self.value == X9ObjectIdentifier.MQV2:
            ret = "Mqv2"
        elif self.value == X9ObjectIdentifier.MQV1:
            ret = "Mqv1"
        elif self.value == X9ObjectIdentifier.SECP384R1:
            ret = "Secp384r1"
        else:
            ret = "Unknown enum value:" + str(self.value)
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.GXIntEnum import GXIntEnum


class X509CertificateType(GXIntEnum):
    """
    x509 Certificate.
    """

    NONE = 0
    OLD_AUTHORITY_KEY_IDENTIFIER = 1
    OLD_PRIMARY_KEY_ATTRIBUTES = 2
    CERTIFICATE_POLICIES = 3
    ORIMARY_KEY_USAGE_RESTRICTION = 4
    SUBJECT_DIRECTORY_ATTRIBUTES = 5
    SUBJECT_KEY_IDENTIFIER = 6
    KEY_USAGE = 7
    PRIVATE_KEY_USAGE_PERIOD = 8
    SUBJECT_ALTERNATIVE_NAME = 9
    ISSUER_ALTERNATIVE_NAME = 10
    BASIC_CONSTRAINTS = 11
    CRL_NUMBER = 12
    REASON_CODE = 13
    HOLD_INSTRUCTION_CODE = 14
    INVALIDITY_DATE = 15
    DELTA_CRL_INDICATOR = 16
    ISSUING_DISTRIBUTION_POINT = 17
    CERTIFICATE_ISSUER = 18
    NAME_CONSTRAINTS = 19
    CRL_DISTRIBUTION_POINTS = 20
    CERTIFICATE_POLICIES2 = 21
    POLICY_MAPPINGS = 22
    AUTHORITY_KEY_IDENTIFIER = 23
    POLICY_CONSTRAINTS = 24
    EXTENDED_KEY_USAGE = 25
    FRESHEST_CRL = 26

    @classmethod
    def valueofString(cls, value):
        return X509CertificateType[value.upper()]

    def __str__(self):
        if self.value == X509CertificateType.NONE:
            ret = "None"
        elif self.value == X509CertificateType.OLD_AUTHORITY_KEY_IDENTIFIER:
            ret = "Old authority key identifier"
        elif self.value == X509CertificateType.OLD_PRIMARY_KEY_ATTRIBUTES:
            ret = "Old primary key attributes"
        elif self.value == X509CertificateType.CERTIFICATE_POLICIES:
            ret = "Certificate policies"
        elif self.value == X509CertificateType.ORIMARY_KEY_USAGE_RESTRICTION:
            ret = "Orimary key usage restriction"
        elif self.value == X509CertificateType.SUBJECT_DIRECTORY_ATTRIBUTES:
            ret = "Subject directory attributes"
        elif self.value == X509CertificateType.SUBJECT_KEY_IDENTIFIER:
            ret = "Subject key identifier"
        elif self.value == X509CertificateType.KEY_USAGE:
            ret = "Key usage"
        elif self.value == X509CertificateType.PRIVATE_KEY_USAGE_PERIOD:
            ret = "Private key usage period"
        elif self.value == X509CertificateType.SUBJECT_ALTERNATIVE_NAME:
            ret = "Subject alternative name"
        elif self.value == X509CertificateType.ISSUER_ALTERNATIVE_NAME:
            ret = "Issuer alternative name"
        elif self.value == X509CertificateType.BASIC_CONSTRAINTS:
            ret = "Basic constraints"
        elif self.value == X509CertificateType.CRL_NUMBER:
            ret = "Crl number"
        elif self.value == X509CertificateType.REASON_CODE:
            ret = "Reason code"
        elif self.value == X509CertificateType.HOLD_INSTRUCTION_CODE:
            ret = "Hold instruction code"
        elif self.value == X509CertificateType.INVALIDITY_DATE:
            ret = "Invalidity date"
        elif self.value == X509CertificateType.DELTA_CRL_INDICATOR:
            ret = "Delta crl indicator"
        elif self.value == X509CertificateType.ISSUING_DISTRIBUTION_POINT:
            ret = "Issuing distribution point"
        elif self.value == X509CertificateType.CERTIFICATE_ISSUER:
            ret = "Certificate issuer"
        elif self.value == X509CertificateType.NAME_CONSTRAINTS:
            ret = "Name constraints"
        elif self.value == X509CertificateType.CRL_DISTRIBUTION_POINTS:
            ret = "Crl distribution points"
        elif self.value == X509CertificateType.CERTIFICATE_POLICIES2:
            ret = "Certificate policies2"
        elif self.value == X509CertificateType.POLICY_MAPPINGS:
            ret = "Policy mappings"
        elif self.value == X509CertificateType.AUTHORITY_KEY_IDENTIFIER:
            ret = "Authority key identifier"
        elif self.value == X509CertificateType.POLICY_CONSTRAINTS:
            ret = "Policy constraints"
        elif self.value == X509CertificateType.EXTENDED_KEY_USAGE:
            ret = "Extended key usage"
        elif self.value == X509CertificateType.FRESHEST_CRL:
            ret = "Freshest crl"
        else:
            ret = "Unknown enum value:" + str(self.value)
        return ret



#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.GXIntFlag import GXIntFlag


class ExtendedKeyUsage(GXIntFlag):
    """
    Extended key usage.
    """

    NONE = 0
    """
    Extended key usage is not used.
    """
    SERVER_AUTH = 1
    """
    Certificate can be used as an TLS server certificate.
    """
    CLIENT_AUTH = 2
    """
    Certificate can be used as an TLS client certificate.
    """



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename:        $HeadURL$
#
#  Version:         $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..GXIntEnum import GXIntEnum

class InterfaceType(GXIntEnum):
    """InterfaceType enumerates the usable types of connection in GuruxDLMS."""
    #pylint: disable=too-few-public-methods

    # General interface type is used for meters that support IEC 62056-46 Data link layer using HDLC protocol.
    HDLC = 0
    # Network interface type is used for meters that support IEC 62056-47 COSEM transport layers for IPv4 networks.
    WRAPPER = 1
    # Plain PDU is returned.
    PDU = 2
    # EN 13757-4/-5 Wireless M-Bus profile is used.
    WIRELESS_MBUS = 3
    # IEC 62056-21 E-Mode is used to initialize communication before moving to HDLC protocol.
    HDLC_WITH_MODE_E = 4
    #PLC Logical link control (LLC) profile is used with IEC 61334-4-32 connectionless LLC sublayer.
    PLC = 5
    # PLC Logical link control (LLC) profile is used with HDLC.
    PLC_HDLC = 6
    # LowPower Wide Area Networks (LPWAN) profile is used.
    LPWAN = 7
    # Wi-SUN FAN mesh network is used.
    WI_SUN = 8
    # OFDM PLC PRIME is defined in IEC 62056-8-4.
    PLC_PRIME = 9
    # EN 13757-2 wired (twisted pair based) M-Bus scheme is used.
    WIRED_MBUS = 10



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.GXIntEnum import GXIntEnum


class PkcsObjectIdentifier(GXIntEnum):
    NONE = 0
    RSA_ENCRYPTION = 1
    MD2WITH_RSA_ENCRYPTION = 2
    MD4WITH_RSA_ENCRYPTION = 3
    MD5WITH_RSA_ENCRYPTION = 4
    SHA1WITH_RSA_ENCRYPTION = 5
    SRSA_OAEP_ENCRYPTION_SET = 6
    ID_RSAES_OAEP = 7
    ID_MGF1 = 8
    ID_P_SPECIFIED = 9
    ID_RSASSA_PSS = 10
    SHA256WITH_RSA_ENCRYPTION = 11
    SHA384WITH_RSA_ENCRYPTION = 12
    SHA512WITH_RSA_ENCRYPTION = 13
    SHA224WITH_RSA_ENCRYPTION = 14
    DH_KEY_AGREE1MENT = 15
    PBE_WITH_MD2AND_DES_CBC = 16
    PBE_WITH_MD2AND_RC2CBC = 17
    PBE_WITH_MD5AND_DES_CBC = 18
    PBE_WITH_MD5AND_RC2CBC = 19
    PBE_WITH_SHA1AND_DES_CBC = 20
    PBE_WITH_SHA1AND_RC2CBC = 21
    ID_PBE_S2 = 22
    ID_PBKDF2 = 23
    DES_EDE3CBC = 24
    RC2CBC = 25
    MD2 = 26
    MD4 = 27
    MD5 = 28
    ID_HMAC_WITH_SHA1 = 29
    ID_HMAC_WITH_SHA224 = 30
    ID_HMAC_WITH_SHA256 = 31
    ID_HMAC_WITH_SHA384 = 32
    ID_HMAC_WITH_SHA512 = 33
    DATA = 34
    SIGNED_DATA = 35
    ENVELOPED_DATA = 36
    SIGNED_AND_ENVELOPED_DATA = 37
    DIGESTED_DATA = 38
    ENCRYPTED_DATA = 39
    PKCS9AT_EMAIL_ADDRESS = 40
    PKCS9AT_UNSTRUCTURED_NAME = 41
    PKCS9AT_CONTENT_TYPE = 42
    PKCS9AT_MESSAGE_DIGEST = 43
    PKCS9AT_SIGNING_TIME = 44
    PKCS9AT_COUNTER_SIGNATURE = 45
    PKCS9AT_CHALLENGE_PASSWORD = 46
    PKCS9AT_UNSTRUCTURED_ADDRESS = 47
    PKCS9AT_EXTENDED_CERTIFICATE_ATTRIBUTES = 48
    PKCS9AT_SIGNING_DESCRIPTION = 49
    PKCS9AT_EXTENSION_REQUEST = 50
    PKCS9AT_SMIME_CAPABILITIES = 51
    ID_SMIME = 52
    PKCS9AT_FRIENDLY_NAME = 53
    PKCS9AT_LOCAL_KEY_ID = 54
    X509CERTIFICATE = 55
    SDSI_CERTIFICATE = 56
    X509CRL = 57
    ID_ALG = 58
    ID_ALG_ESDH = 59
    ID_ALG_CMS3DES_WRAP = 60
    ID_ALG_CMS_RC2WRAP = 61
    ID_ALG_PWRI_KEK = 62
    ID_ALG_SSDH = 63
    ID_RSA_KEM = 64
    PREFER_SIGNED_DATA = 65
    CANNOT_DECRYPT_ANY = 66
    SMIME_CAPABILITIES_VERSIONS = 67
    ID_AA_RECEIPT_REQUEST = 68
    ID_CT_AUTH_DATA = 69
    ID_CT_TST_INFO = 70
    ID_CT_COMPRESSED_DATA = 71
    ID_CT_AUTH_ENVELOPED_DATA = 72
    ID_CT_TIMESTAMPED_DATA = 73
    ID_CTI_ETS_PROOF_OF_ORIGIN = 74
    ID_CTI_ETS_PROOF_OF_RECEIPT = 75
    ID_CTI_ETS_PROOF_OF_DELIVERY = 76
    ID_CTI_ETS_PROOF_OF_SENDER = 77
    ID_CTI_ETS_PROOF_OF_APPROVAL = 78
    ID_CTI_ETS_PROOF_OF_CREATION = 79
    ID_AA_CONTENT_HINT = 80
    ID_AA_MSG_SIG_DIGEST = 81
    ID_AA_CONTENT_REFERENCE = 82
    ID_AA_ENCRYP_KEY_PREF = 83
    ID_AA_SIGNING_CERTIFICATE = 84
    ID_AA_SIGNING_CERTIFICATE_V2 = 85
    ID_AA_CONTENT_IDENTIFIER = 86
    ID_AA_SIGNATURE_TIME_STAMP_TOKEN = 87
    ID_AA_ETS_SIG_POLICY_ID = 88
    ID_AA_ETS_COMMITMENT_TYPE = 89
    ID_AA_ETS_SIGNER_LOCATION = 90
    ID_AA_ETS_SIGNER_ATTR = 91
    ID_AA_ETS_OTHER_SIG_CERT = 92
    ID_AA_ETS_CONTENT_TIMESTAMP = 93
    ID_AA_ETS_CERTIFICATE_REFS = 94
    ID_AA_ETS_REVOCATION_REFS = 95
    ID_AA_ETS_CERT_VALUES = 96
    ID_AA_ETS_REVOCATION_VALUES = 97
    ID_AA_ETS_ESC_TIME_STAMP = 98
    ID_AA_ETS_CERT_CRL_TIMESTAMP = 99
    ID_AA_ETS_ARCHIVE_TIMESTAMP = 100
    ID_SPQ_ETS_URI = 101
    ID_SPQ_ETS_U_NOTICE = 102
    KEY_BAG = 103
    PKCS8SHROUDED_KEY_BAG = 104
    CERT_BAG = 105
    CRL_BAG = 106
    SECRET_BAG = 107
    SAFE_CONTENTS_BAG = 108
    PBE_WITH_SHA_AND128BIT_RC4 = 109
    PBE_WITH_SHA_AND40BIT_RC4 = 110
    PBE_WITH_SHA_AND3KEY_TRIPLE_DES_CBC = 111
    PBE_WITH_SHA_AND2KEY_TRIPLE_DES_CBC = 112
    PBE_WITH_SHA_AND128BIT_RC2CBC = 113
    PBEWITH_SHA_AND40BIT_RC2CBC = 114



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import datetime
from .GXDateTime import GXDateTime
from .enums import DateTimeSkips

class GXDate(GXDateTime):
    def __init__(self, value=None, pattern=None):
        """
        Constructor.

        pattern: Date-time pattern that is used when value is a string.
        value: Date value.
        """
        GXDateTime.__init__(self, value, pattern)
        self.skip |= DateTimeSkips.HOUR
        self.skip |= DateTimeSkips.MINUTE
        self.skip |= DateTimeSkips.SECOND
        self.skip |= DateTimeSkips.MILLISECOND

    def _remove(self, format_):
        format_ = GXDateTime._remove_(format_, "%H", True)
        format_ = GXDateTime._remove_(format_, "%-H", True)
        format_ = GXDateTime._remove_(format_, "%I", True)
        format_ = GXDateTime._remove_(format_, "%-I", True)
        format_ = GXDateTime._remove_(format_, "%M", True)
        format_ = GXDateTime._remove_(format_, "%-M", True)
        format_ = GXDateTime._remove_(format_, "%S", True)
        format_ = GXDateTime._remove_(format_, "%-S", True)
        str_ = datetime.datetime.now().strftime('%p')
        format_ = GXDateTime._remove_(format_, str_, True)
        return format_



#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------

from gurux_dlms.GXIntFlag import GXIntFlag


class KeyUsage(GXIntFlag):
    """
    Key Usage.
    """

    NONE = 0
    """
    Key is not used.
    """
    DIGITAL_SIGNATURE = 0x1
    """
    Digital signature.
    """
    NON_REPUDIATION = 0x2
    """
    Non Repudiation.
    """
    KEY_ENCIPHERMENT = 0x4
    """
    Key encipherment.
    """
    DATA_ENCIPHERMENT = 0x8
    """
    Data encipherment.
    """
    KEY_AGREEMENT = 0x10
    """
    Key agreement.
    """
    KEY_CERT_SIGN = 0x20
    """
    Used with CA certificates when the subject public key is used to verify a signature on certificates.
    """
    CRL_SIGN = 0x40
    """
    Used when the subject public key is to verify a signature.
    """
    ENCIPHER_ONLY = 0x80
    """
    Encipher only.
    """
    DECIPHER_ONLY = 0x100
    """
    Decipher only.
    """



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..GXBitString import GXBitString
from ..asn.GXAsn1Sequence import GXAsn1Sequence
from ..GXByteBuffer import GXByteBuffer


# pylint: disable=too-few-public-methods, import-outside-toplevel
class GXAsn1PublicKey:
    """
    ASN1 Public key.
    """

    def __init__(self, key=None):
        """
        Constructor.
        - key: bytes or GXBitString
        """
        self.value = None
        if key is None:
            return

        if isinstance(key, GXBitString):
            from gurux_dlms.asn.GXAsn1Converter import GXAsn1Converter

            seq = GXAsn1Converter.fromByteArray(key.value)
            if not isinstance(seq, GXAsn1Sequence):
                raise ValueError("Invalid ASN1 sequence")
            self._init(GXAsn1Converter.toByteArray([seq[0], seq[1]]))
        elif isinstance(key, (bytes, bytearray)):
            self._init(key)
        else:
            raise TypeError("Invalid argument type for key")

    def _init(self, key: bytes):
        if key is None or len(key) != 270:
            raise ValueError("data")
        self.value = bytearray(key)  # Kopioidaan key ? value

    def __str__(self):
        """
        Returns the public key as a hex string.
        """
        if self.value is None:
            return ""
        return GXByteBuffer.hex(self.value, False)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.CryptoKeyType import CryptoKeyType
from .enums.Security import Security
from .objects.enums.SecuritySuite import SecuritySuite
from .objects.enums.SecurityPolicy import SecurityPolicy
from .objects.enums.CertificateType import CertificateType


# pylint: disable=too-many-instance-attributes,too-few-public-methods
class GXCryptoKeyParameter:
    def __init__(self):
        self.command = 0
        # Crypto key type.
        self.keyType = CryptoKeyType.BLOCK_CIPHER
        # Is data encrypted or decrypted.
        self.encrypt = False
        # Encrypted data.
        self.encrypted = None
        # Decrypted data.
        self.plainText = None
        # Used security.
        self.security = Security.NONE
        # Used security suite.
        self.securitySuite = SecuritySuite.SUITE_0
        # Used security policy.
        self.securityPolicy = SecurityPolicy.NOTHING
        # Used certificate type.
        self.certificateType = CertificateType.DIGITAL_SIGNATURE
        # System title
        self.systemTitle = None
        # Recipient system title.
        self.recipientSystemTitle = None
        # Block cipher key.
        self.blockCipherKey = None
        # Authentication key.
        self.authenticationKey = None
        # Frame counter. Invocation counter.
        self.invocationCounter = 0
        # Transaction Id.
        self.transactionId = None
        # key to used to encrypt the data.
        self.privateKey = None
        # Public key to used to decrypt the data.
        self.publicKey = None



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                $Date$
#                $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import gettext
import os
from collections.abc import Iterable

# This class is used to localize the texts.
class _GXLocalizer:
    #Used language.
    __lang = None

    # pylint: disable=bare-except, broad-exception-caught
    # Return localized value.
    @staticmethod
    def gettext(value):
        try:
            if _GXLocalizer.__lang:
                return _GXLocalizer.__lang.gettext(value)
        except Exception:
            pass
        #Return default value if language is not available.
        return value

    # pylint: disable=bare-except
    @staticmethod
    def init(lang_code):
        try:
            if isinstance(lang_code, Iterable) and not isinstance(lang_code, (str, bytes)):
                if "_" in lang_code[0]:
                    lang_code = lang_code[0].split('_')[0]
            locale_path = os.path.join(os.path.dirname(__file__), "locale")
            _GXLocalizer.__lang = gettext.translation(
                "gurux_dlms", localedir=locale_path, languages=[lang_code], fallback=True
            )
            _GXLocalizer.__lang.install()
        except Exception:
            pass



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums import TokenDelivery, TokenStatusCode
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSTokenGateway(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSTokenGateway
    """

    def __init__(self, ln="0.0.19.40.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        super(GXDLMSTokenGateway, self).__init__(ObjectType.TOKEN_GATEWAY, ln, sn)
        # Descriptions.
        self.descriptions = []
        # Token Delivery method.
        self.deliveryMethod = TokenDelivery.LOCAL
        # Token status code.
        self.statusCode = TokenStatusCode.FORMAT_OK
        # Token.
        self.token = []
        # Time.
        self.time = None
        # Token data value.
        self.dataValue = None

    def getValues(self):
        return [
            self.logicalName,
            self.token,
            self.time,
            self.descriptions,
            self.deliveryMethod,
            [self.statusCode, self.dataValue],
        ]

    #
    # Returns collection of attributes to read.  If attribute is static
    #      and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  Token
        if all_ or self.canRead(2):
            attributes.append(2)
        #  Time
        if all_ or self.canRead(3):
            attributes.append(3)
        #  Description
        if all_ or self.canRead(4):
            attributes.append(4)
        #  DeliveryMethod
        if all_ or self.canRead(5):
            attributes.append(5)
        #  Status
        if all_ or self.canRead(6):
            attributes.append(6)
        return attributes

    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 6

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Token"),\
            _GXLocalizer.gettext("Time"),\
            _GXLocalizer.gettext("Description"),\
            _GXLocalizer.gettext("Delivery method"),\
            _GXLocalizer.gettext("Status"))

    def getMethodNames(self):
        return (_GXLocalizer.gettext("Enter"),)

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.OCTET_STRING
        elif index == 3:
            ret = DataType.OCTET_STRING
        elif index == 4:
            ret = DataType.ARRAY
        elif index == 5:
            ret = DataType.ENUM
        elif index == 6:
            ret = DataType.STRUCTURE
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        # pylint: disable=bad-option-value,redefined-variable-type
        bb = None
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.token
        elif e.index == 3:
            ret = self.time
        elif e.index == 4:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if not self.descriptions:
                bb.setUInt8(0)
            else:
                bb.setUInt8(len(self.descriptions))
                for it in self.descriptions:
                    bb.setUInt8(DataType.OCTET_STRING)
                    bb.setUInt8(int(len(it)))
                    bb.set(it)
            ret = bb
        elif e.index == 5:
            ret = self.deliveryMethod
        elif e.index == 6:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.STRUCTURE)
            bb.setUInt8(2)
            _GXCommon.setData(settings, bb, DataType.ENUM, self.statusCode)
            _GXCommon.setData(settings, bb, DataType.BITSTRING, self.dataValue)
            ret = bb
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
        return ret

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.token = e.value
        elif e.index == 3:
            self.time = _GXCommon.changeType(settings, e.value, DataType.DATETIME)
        elif e.index == 4:
            self.descriptions = []
            if e.value:
                for it in e.value:
                    self.descriptions.append(it)
        elif e.index == 5:
            self.deliveryMethod = e.value
        elif e.index == 6:
            self.statusCode = e.value[0]
            self.dataValue = str(e.value[1])
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.token = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Token"))
        self.time = reader.readElementContentAsDateTime("Time")
        self.descriptions = []
        if reader.isStartElement("Descriptions", True):
            while reader.isStartElement("Item", True):
                self.descriptions.append(reader.readElementContentAsString("Name"))
            reader.readEndElement("Descriptions")
        self.deliveryMethod = reader.readElementContentAsInt("DeliveryMethod")
        self.statusCode = reader.readElementContentAsInt("Status")
        self.dataValue = reader.readElementContentAsString("Data")

    def save(self, writer):
        writer.writeElementString("Token", GXByteBuffer.hex(self.token, False))
        writer.writeElementString("Time", self.time)
        if self.descriptions:
            writer.writeStartElement("Descriptions")
            for it in self.descriptions:
                writer.writeStartElement("Item")
                writer.writeElementString("Name", it)
                writer.writeEndElement()
            writer.writeEndElement()
        writer.writeElementString("DeliveryMethod", int(self.deliveryMethod))
        writer.writeElementString("Status", int(self.statusCode))
        writer.writeElementString("Data", self.dataValue)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import sys
import re
from .enums.CertificateVersion import CertificateVersion
from .enums.HashAlgorithm import HashAlgorithm
from .enums.PkcsObjectIdentifier import PkcsObjectIdentifier
from .enums.X9ObjectIdentifier import X9ObjectIdentifier
from ..ecdsa.GXPublicKey import GXPublicKey
from ..GXInt8 import GXInt8
from ..GXByteBuffer import GXByteBuffer
from ..GXBitString import GXBitString
from ..ecdsa.GXDLMSCertificateException import GXDLMSCertificateException
from .GXAsn1Sequence import GXAsn1Sequence
from .GXAsn1Converter import GXAsn1Converter
from .PkcsObjectIdentifierConverter import PkcsObjectIdentifierConverter
from .GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier
from .enums.PkcsType import PkcsType
from ..ecdsa.enums.Ecc import Ecc
from .GXAsn1Context import GXAsn1Context
from .GXAsn1Integer import GXAsn1Integer
from ..internal._GXCommon import _GXCommon
from .HashAlgorithmConverter import HashAlgorithmConverter
from .X9ObjectIdentifierConverter import X9ObjectIdentifierConverter
from ..ecdsa.GXEcdsa import GXEcdsa
from ..objects.enums.CertificateType import CertificateType
from ..asn.enums.KeyUsage import KeyUsage
from .enums.ExtendedKeyUsage import ExtendedKeyUsage
from .GXx509Certificate import GXx509Certificate
from ..GXArray import GXArray

if sys.version_info >= (3, 0):
    import urllib.request


# pylint: disable=too-many-instance-attributes
class GXPkcs10:
    """
    Pkcs10 Certificate Signing Request.
    """

    __rawData = bytearray()
    """
    Loaded PKCS #10 certificate as a raw data.
    """

    __version = CertificateVersion.VERSION1
    """
    Certificate version.
    """

    __subject = ""
    """
    Subject.
    """

    __attributes = []
    """
    Collection of attributes providing additional information about the subject of the certificate.
    """

    __algorithm = X9ObjectIdentifier.ID_EC_PUBLIC_KEY
    """
    Algorithm.
    """

    __publicKey = None
    """
    Subject public key.
    """

    __signatureAlgorithm = HashAlgorithm.NONE
    """
    Signature algorithm.
    """

    __signatureParameters = None
    """
    Signature parameters.
    """

    __signature = bytearray()
    """
    Signature.
    """

    @property
    def version(self):
        """
        Certificate version.
        """
        return self.__version

    @property
    def subject(self):
        """
        Subject.
        """
        return self.__subject

    @property
    def attributes(self):
        """
        Collection of attributes providing additional information about the subject of the certificate.
        """
        return self.__attributes

    @property
    def algorithm(self):
        """
        Algorithm.
        """
        return self.__algorithm

    @property
    def publicKey(self):
        """
        Subject public key.
        """
        return self.__publicKey

    @property
    def signatureAlgorithm(self):
        """
        Signature algorithm.
        """
        return self.__signatureAlgorithm

    @property
    def signatureParameters(self):
        """
        Signature parameters.
        """
        return self.__signatureParameters

    @property
    def signature(self):
        """
        Signature.
        """
        return self.__signature

    @property
    def encoded(self):
        if self.__rawData:
            return self.__rawData
        if not self.signature:
            raise ValueError("Sign first.")

        # Certification request info.
        # subject Public key info.
        sa = GXAsn1ObjectIdentifier(
            HashAlgorithmConverter.getString(self.__signatureAlgorithm)
        )
        value = GXArray(
            (self.__getData(), GXArray((sa,)), GXBitString(self.signature, 0))
        )
        return GXAsn1Converter.toByteArray(value)

    def __init__(self, value=None):
        """
        Constructor.
        """
        if value:
            self.__init(value)

    # pylint: disable=too-many-locals, protected-access
    def __init(self, data):
        self.__rawData = data
        seq = GXAsn1Converter.fromByteArray(data)
        if len(seq) < 3:
            raise ValueError("Wrong number of elements in sequence.")
        if not isinstance(seq[0], GXAsn1Sequence):
            type_ = GXAsn1Converter.__getCertificateType(data, seq)
            if type_ == PkcsType.PKCS8:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is PKCS 8, not PKCS 10."
                )
            if type_ == PkcsType.X509CERTIFICATE:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is PKCS x509 certificate, not PKCS 10."
                )
            raise GXDLMSCertificateException("Invalid Certificate Version.")

        reqInfo = seq[0]
        self.__version = CertificateVersion(reqInfo[0])
        self.__subject = GXAsn1Converter.getSubject(reqInfo[1])
        # subject Public key info.
        subjectPKInfo = reqInfo[2]
        if len(reqInfo) > 3:
            # PkcsObjectIdentifier
            for it in reqInfo[3]:
                values = []
                for _, v in it[1]:
                    values.append(v)
                self.attributes.append(
                    (PkcsObjectIdentifierConverter.fromString(str(it[0])), values)
                )

        tmp = GXAsn1Sequence(subjectPKInfo[0])
        self.__algorithm = X9ObjectIdentifierConverter.fromString(str(tmp[0]))
        if self.__algorithm != X9ObjectIdentifier.ID_EC_PUBLIC_KEY:
            algorithm = self.algorithm
            if self.algorithm == X9ObjectIdentifier.NONE:
                algorithm = PkcsObjectIdentifierConverter.fromString(str(tmp[0]))
                if algorithm == PkcsObjectIdentifier.NONE:
                    algorithm = str(tmp[0])
            raise ValueError("Invalid PKCS #10 certificate algorithm. " + algorithm)

        self.__publicKey = GXPublicKey.fromRawBytes(subjectPKInfo[1].value)
        GXEcdsa.validate(self.publicKey)
        # signatureAlgorithm
        sign = seq[1]
        self.__signatureAlgorithm = HashAlgorithmConverter.fromString(str(sign[0]))
        if not self.__signatureAlgorithm in (
            HashAlgorithm.SHA256WITH_ECDSA,
            HashAlgorithm.SHA384WITH_ECDSA,
        ):
            raise GXDLMSCertificateException(
                "Invalid signature algorithm. " + str(sign[0])
            )
        if len(sign) != 1:
            self.signatureParameters = sign[1]
        # signature
        # Get raw data
        tmp2 = GXByteBuffer()
        tmp2.set(data)
        GXAsn1Converter._getNext(tmp2)
        tmp2.size = tmp2.position
        tmp2.position = 1
        _GXCommon.getObjectCount(tmp2)
        # Get signature.
        self.__signature = seq[2].value
        e = GXEcdsa(self.publicKey)
        tmp3 = GXAsn1Converter.fromByteArray(self.signature)
        bb = GXByteBuffer()
        if self.__signatureAlgorithm == HashAlgorithm.SHA256WITH_ECDSA:
            size = 32
        else:
            size = 48
        # Some implementations might add extra byte. It must removed.
        if tmp3[0].length == size:
            start = 0
        else:
            start = 1

        bb.set(tmp3[0].toArray(), start, size)
        if tmp3[1].length == size:
            start = 0
        else:
            start = 1
        bb.set(tmp3[1].toArray(), start, size)
        if not e.verify(bb.array(), tmp2.subArray(tmp2.position, tmp2.available())):
            raise ValueError("Invalid Signature.")

    def __getData(self):
        if self.publicKey.scheme == Ecc.P256:
            alg = GXAsn1ObjectIdentifier("1.2.840.10045.3.1.7")
        else:
            alg = GXAsn1ObjectIdentifier("1.3.132.0.34")

        subjectPKInfo = GXBitString(self.publicKey.rawValue, 0)
        tmp = GXArray([GXAsn1ObjectIdentifier("1.2.840.10045.2.1"), alg])
        attributes = GXAsn1Context()
        for k, v in self.attributes:
            s = GXAsn1Sequence()
            s.append(GXAsn1ObjectIdentifier(PkcsObjectIdentifierConverter.getString(k)))
            # Convert object array to list.
            values = v
            s.append((values, None))
            attributes.append(s)
        return GXArray(
            [
                GXInt8(self.version),
                GXAsn1Converter.encodeSubject(self.subject),
                GXArray([tmp, subjectPKInfo]),
                attributes,
            ]
        )

    def sign(self, key, hashAlgorithm):
        """
        Sign

            Parameters:
                key: Private key.
                hashAlgorithm: Used algorithm for signing.
        """
        data = GXAsn1Converter.toByteArray(self.__getData())
        e = GXEcdsa(key)
        self.__signatureAlgorithm = hashAlgorithm
        bb = e.sign(data)
        if self.__signatureAlgorithm == HashAlgorithm.SHA256WITH_ECDSA:
            size = 32
        else:
            size = 48
        tmp = GXArray((GXAsn1Integer(bb[0:size]), GXAsn1Integer(bb[size:])))
        self.__signature = GXAsn1Converter.toByteArray(tmp)

    @classmethod
    def fromHexString(cls, data):
        """
        Create PKCS 10 from hex string.

            Parameters:
                data: Hex string.

            Returns:
                PKCS 10
        """
        cert = GXPkcs10()
        cert.__init(GXByteBuffer.hexToBytes(data))
        return cert

    @classmethod
    def fromPem(cls, data):
        """
        Create x509Certificate from PEM string.

            Parameters:
                data: PEM string.
        """
        data.replace("\r\n", "\n")
        START = "CERTIFICATE REQUEST-----\n"
        END = "-----END"
        data = data.replace("\r\n", "\n")
        start = data.find(START)
        if start == -1:
            raise ValueError("Invalid PEM file.")
        data = data[start + len(START) :]
        end = data.find(END)
        if end == -1:
            raise ValueError("Invalid PEM file.")
        return cls.fromDer(data[0:end].strip())

    @classmethod
    def fromDer(cls, der):
        """
        Create x509Certificate from DER Base64 encoded string.

            Parameters:
                der: Base64 DER string.

            Returns:
        """
        der = der.replace("\r\n", "")
        der = der.replace("\n", "")
        cert = GXPkcs10()
        cert.__init(_GXCommon.fromBase64(der))
        return cert

    def toString(self):
        bb = "PKCS #10 certificate request:"
        bb += "\r\n"
        bb += "Version: " + str(self.version)
        bb += "\r\n"
        bb += "Subject: " + self.subject
        bb += "\r\n"
        bb += "Algorithm: " + str(self.algorithm)
        bb += "\r\n"
        bb += "Public Key: "
        if self.publicKey:
            bb += str(self.publicKey)
        bb += "\r\n"
        bb += "Signature algorithm: " + str(self.__signatureAlgorithm)
        bb += "\r\n"
        if self.signatureParameters:
            bb += "Signature parameters: "
            bb += str(self.signatureParameters)
            bb += "\r\n"

        bb += "Signature: " + GXByteBuffer.toHex(self.signature, False)
        bb += "\r\n"
        return bb

    # pylint: disable=unused-private-member
    @classmethod
    def createCertificateSigningRequest(cls, kp, subject):
        """
        Create Certificate Signing Request.

            Parameters:
                kp: KeyPair
                subject: Subject.

            Returns:
                Created GXPkcs10.
        """
        if not subject or subject.find("CN=") == -1:
            raise ValueError("subject")

        pkc10 = GXPkcs10()
        pkc10.__algorithm = X9ObjectIdentifier.ID_EC_PUBLIC_KEY
        pkc10.__publicKey = kp[0]
        pkc10.__subject = subject
        if pkc10.__publicKey.scheme == Ecc.P256:
            ha = HashAlgorithm.SHA256WITH_ECDSA
        else:
            ha = HashAlgorithm.SHA384WITH_ECDSA
        pkc10.sign(kp[1], ha)
        return pkc10

    @classmethod
    def getCertificate(cls, address, certifications):
        """
        Ask Gurux certificate server to generate the new certificate.

            Parameters:
                address: Certificate server address.
                certifications: List of certification requests.

            Returns:
                Generated certificate(s).
        """
        usage = ""
        for it in certifications:
            if usage:
                usage += ", "
            usage += '{"KeyUsage":'
            if it.certificateType == CertificateType.DIGITAL_SIGNATURE:
                usage += str(int(KeyUsage.DIGITAL_SIGNATURE))
            elif it.certificateType == CertificateType.KEY_AGREEMENT:
                usage += str(int(KeyUsage.KEY_AGREEMENT))
            elif it.certificateType == CertificateType.TLS:
                usage += str(int(KeyUsage.DIGITAL_SIGNATURE | KeyUsage.KEY_AGREEMENT))
            else:
                raise ValueError("Invalid type.")

            if it.extendedKeyUsage != ExtendedKeyUsage.NONE:
                usage += ', "ExtendedKeyUsage":'
                usage += str(int(it.extendedKeyUsage))
            usage += ', "CSR":"'
            usage += it.certificate.toDer()
            usage += '"}'

        der = '{"Certificates":[' + usage + "]}"
        req = urllib.request.Request(address, der.encode("utf-8"), method="POST")
        req.add_header("Content-Type", "application/json")
        with urllib.request.urlopen(req) as response:
            result = response.read().decode("utf-8")
            pos = result.find("[")
            if pos == -1:
                raise ValueError("Certificates are missing.")
            result = result[pos + 2 :]
            pos = result.find("]")
            if pos == -1:
                raise ValueError("Certificates are missing.")

            result = result[0 : pos - 1]
            certs = []
            tmp = re.split(r'[";,]', result)
            tmp = [p for p in tmp if p.strip()]
            pos = 0
            for it in tmp:
                # Ignore empty.
                if it:
                    x509 = GXx509Certificate.fromDer(it)
                    if (
                        certifications[pos].certificate.publicKey.rawValue
                        != x509.publicKey.rawValue
                    ):
                        raise ValueError(
                            "Create certificate signingRequest generated wrong public key."
                        )
                    pos += 1
                    certs.append(x509)
        return certs

    @classmethod
    def load(cls, path):
        """
        Load Pkcs10 Certificate Signing Request from the PEM (.csr) file.

            Parameters:
                path: File path.

            Returns:
                Created GXPkcs10 object.
        """

        with open(path, "r", encoding="utf-8") as file:
            return cls.fromPem(file.read())

    def save(self, path):
        """
        Save Pkcs #10 Certificate Signing Request to PEM file.

            Parameters:
                path: File path.
        """

        with open(path, "w", encoding="utf-8") as file:
            file.write(self.toPem())

    def toPem(self):
        """
        Pkcs #10 Certificate Signing Request in DER format.
            Returns:
                Public key as in PEM string.
        """
        sb = "-----BEGIN CERTIFICATE REQUEST-----\r"
        sb += self.toDer()
        sb += "-----END CERTIFICATE REQUEST-----\r"
        return sb

    def toDer(self):
        """
        Pkcs #10 Certificate Signing Request in DER format.
            Returns:
                Public key as in PEM string.
        """
        return _GXCommon.toBase64(self.encoded)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.Ecc import Ecc
from .GXEccPoint import GXEccPoint


class GXCurve:
    """
    ECC x and y points in the curve.
    """

    __a = None
    """
    ECC curve a value.
    """

    __p = None
    """
    ECC curve p value.
    """

    __b = None
    """
    ECC curve b parameter.
    """

    __g = None
    """
    x and y-coordinate of base point G
    """

    __n = None
    """
    Order of point G in ECC curve.
    """

    @property
    def a(self):
        """
        ECC curve a value.
        """
        return self.__a

    @property
    def p(self):
        """
        ECC curve p value.
        """
        return self.__p

    @property
    def b(self):
        """
        ECC curve b parameter.
        """
        return self.__b

    @property
    def g(self):
        """
        x and y-coordinate of base point G
        """
        return self.__g

    @property
    def n(self):
        """
        Order of point G in ECC curve.
        """
        return self.__n

    def __init__(self, scheme):
        """
        Constructor.

            Parameters:
                a: ECC curve a value.
                b: ECC curve b parameter.
                p: ECC curve p value.
                g: x and y-coordinate of base point G
                n: Order of point G in ECC curve.
        """

        if scheme == Ecc.P256:
            self.__a = (
                0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFC
            )
            self.__g = GXEccPoint(
                0x6B17D1F2E12C4247F8BCE6E563A440F277037D812DEB33A0F4A13945D898C296,
                0x4FE342E2FE1A7F9B8EE7EB4A7C0F9E162BCE33576B315ECECBB6406837BF51F5,
            )
            self.__n = (
                0xFFFFFFFF00000000FFFFFFFFFFFFFFFFBCE6FAADA7179E84F3B9CAC2FC632551
            )
            self.__p = (
                0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF
            )
            self.__b = (
                0x5AC635D8AA3A93E7B3EBBD55769886BC651D06B0CC53B0F63BCE3C3E27D2604B
            )
        elif scheme == Ecc.P384:
            self.__a = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFF0000000000000000FFFFFFFC
            self.__g = GXEccPoint(
                0xAA87CA22BE8B05378EB1C71EF320AD746E1D3B628BA79B9859F741E082542A385502F25DBF55296C3A545E3872760AB7,
                0x3617DE4A96262C6F5D9E98BF9292DC29F8F41DBD289A147CE9DA3113B5F0B8C00A60B1CE1D7E819D7A431D7C90EA0E5F,
            )
            self.__n = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFC7634D81F4372DDF581A0DB248B0A77AECEC196ACCC52973
            self.__p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFF0000000000000000FFFFFFFF
            self.__b = 0xB3312FA7E23EE7E4988E056BE3F82D19181D9C6EFE8141120314088F5013875AC656398D8A2ED19D2A85C8EDD3EC2AEF
        else:
            raise ValueError("Invalid scheme.")
